//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file factor_inertial_kinematics.h
 *
 *  Created on: Feb 17, 2020
 *      \author: mfourmy
 */

#ifndef FACTOR_INERTIAL_KINEMATICS_H_
#define FACTOR_INERTIAL_KINEMATICS_H_

#include "core/factor/factor_autodiff.h"
#include "bodydynamics/feature/feature_inertial_kinematics.h"

namespace wolf
{

WOLF_PTR_TYPEDEFS(FactorInertialKinematics);

class FactorInertialKinematics : public FactorAutodiff<FactorInertialKinematics, 9,3,4,3, 3,3,3,3,6>
{
    public:
        FactorInertialKinematics(const FeatureBasePtr&   _feat,
                                 const StateBlockPtr&    _sb_bias_imu,
                                 const ProcessorBasePtr& _processor_ptr,
                                 bool                    _apply_loss_function,
                                 FactorStatus            _status = FAC_ACTIVE);

        ~FactorInertialKinematics() override { /* nothing */ }

        // to keep track of the imu bias, not the standard way to proceed
        StateBlockPtr sb_bias_imu_;

       /*
        Factor state blocks:
        _pb: W_p_WB -> base position in world frame
        _qb: W_q_B  -> base orientation in world frame
        _vb: W_v_WB -> base velocity in world frame
        _c: W_p_WC -> COM position in world frame
        _cd: W_v_WC -> COM velocity in world frame
        _Lc: W_Lc   -> angular momentum at the COM in body frame
        _bp: B_b_BC -> bias on relative body->COM position kinematic measurement
        _baw: 6d Imu bias containing [acc bias, gyro bias]
        */
        template<typename T>
        bool operator () (
            const T* const _pb,
            const T* const _qb,
            const T* const _vb,
            const T* const _c,
            const T* const _cd,
            const T* const _Lc,
            const T* const _bp,
            const T* const _baw,
            T* _res) const;

        Vector9d residual() const;
        double cost() const;

};

} /* namespace wolf */


namespace wolf {

FactorInertialKinematics::FactorInertialKinematics(
                            const FeatureBasePtr&   _feat,
                            const StateBlockPtr&    _sb_bias_imu,
                            const ProcessorBasePtr& _processor_ptr,
                            bool                    _apply_loss_function,
                            FactorStatus            _status) :
    FactorAutodiff("FactorInertialKinematics",
                   TOP_GEOM,
                   _feat,
                    nullptr,              // _frame_other_ptr
                    nullptr,              // _capture_other_ptr
                    nullptr,              // _feature_other_ptr
                    nullptr,              // _landmark_other_ptr
                    _processor_ptr,
                    _apply_loss_function,
                    _status,
                    _feat->getFrame()->getP(),
                    _feat->getFrame()->getO(),
                    _feat->getFrame()->getV(),
                    _feat->getFrame()->getStateBlock('C'),  // COM position
                    _feat->getFrame()->getStateBlock('D'),  // COM velocity (bad name)
                    _feat->getFrame()->getStateBlock('L'),  // Angular momentum
                    _feat->getCapture()->getStateBlock('I'),  // body-to-COM relative position bias
                    _sb_bias_imu
    ),
    sb_bias_imu_(_sb_bias_imu)
{
}

Vector9d FactorInertialKinematics::residual() const{
    Vector9d res;
    double * pb, * qb, * vb, * c, * cd, * Lc, * bp, * baw;
    pb = getFrame()->getP()->getState().data();
    qb = getFrame()->getO()->getState().data();
    vb = getFrame()->getV()->getState().data();
    c = getFrame()->getStateBlock('C')->getState().data();
    cd = getFrame()->getStateBlock('D')->getState().data();
    Lc = getFrame()->getStateBlock('L')->getState().data();
    bp = getCapture()->getStateBlock('I')->getState().data();
    baw = sb_bias_imu_->getState().data();

    operator() (pb, qb, vb, c, cd, Lc, bp, baw, res.data());
    // std::cout << "res: " << res.transpose() << std::endl;
    return res;
}

double FactorInertialKinematics::cost() const{
    return residual().squaredNorm();
}

template<typename T>
bool FactorInertialKinematics::operator () (
                    const T* const _pb,
                    const T* const _qb,
                    const T* const _vb,
                    const T* const _c,
                    const T* const _cd,
                    const T* const _Lc,
                    const T* const _bp,
                    const T* const _baw,
                    T* _res) const
{
    // State variables instanciation
    Map<const Matrix<T,3,1> > pb(_pb);
    Map<const Quaternion<T> > qb(_qb);
    Map<const Matrix<T,3,1> > vb(_vb);
    Map<const Matrix<T,3,1> > c(_c);
    Map<const Matrix<T,3,1> > cd(_cd);
    Map<const Matrix<T,3,1> > Lc(_Lc);
    Map<const Matrix<T,3,1> > bp(_bp);
    Map<const Matrix<T,6,1> > baw(_baw);

    Map<Matrix<T,9,1> > res(_res);

    auto feat = std::static_pointer_cast<const FeatureInertialKinematics>(getFeature());

    // Measurements retrieval
    Map<const Vector3d> pBC_m(getMeasurement().data()    );      // B_p_BC
    Map<const Vector3d> vBC_m(getMeasurement().data() + 3);  // B_v_BC
    Map<const Vector3d> w_m  (getMeasurement().data() + 6);      // B_wB

    // Error variable instanciation
    Matrix<T, 9, 1> err;
    Map<Matrix<T, 3, 1> > err_c (err.data()    );
    Map<Matrix<T, 3, 1> > err_cd(err.data() + 3);
    Map<Matrix<T, 3, 1> > err_Lc(err.data() + 6);

    err_c  = pBC_m - (qb.conjugate()*(c - pb) + bp);
    err_cd = vBC_m + (w_m - baw.tail(3)).cross(pBC_m - bp) - qb.conjugate()*(cd - vb);
    err_Lc = feat->getBIq() * (w_m - baw.tail(3)) + feat->getBLcm() -  qb.conjugate()*Lc;

    res = getMeasurementSquareRootInformationUpper() * err;

    return true;
}

} // namespace wolf

#endif /* FACTOR__INERTIAL_KINEMATICS_H_ */
