//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * force_torque_inertial_delta_tools.h
 *
 *  Created on: Aug 8, 2021
 *      Author: jsola
 */

#ifndef MATH_FORCE_TORQUE_INERTIAL_DELTA_TOOLS_H_
#define MATH_FORCE_TORQUE_INERTIAL_DELTA_TOOLS_H_

#include <core/common/wolf.h>
#include <core/math/rotations.h>
#include <core/state_block/state_composite.h>

/*
 * Most functions in this file are explained in the document:
 *
 *     https://www.overleaf.com/project/6107c87f581beb50f732ce5f
 *
 * The functions relate manipulations of Delta motion magnitudes used for Force torque and inertial sensors
 * pre-integration. See the following figure for the flow of calculations of the deltas
 *
 *     https://gitlab.iri.upc.edu/mobile_robotics/wolf_projects/wolf_lib/plugins/bodydynamics/-/blob/12-add-functionality-for-aerial-vehicles/doc/figures/dyn-model.pdf
 *
 * The Delta, of type VectorXd(19) and size 19, is defined as
 *     Delta = [Dpi, Dvi, Dpd, Dvd, DL, Dq]
 * with
 *     Dpi : position delta as integrated by IMU
 *     Dvi : velocity delta as integrated by IMU
 *     Dpd : position delta as integrated by force-torque
 *     Dvd : velocity delta as integrated by force-torque
 *     DL  : CoM Angular momentum delta as integrated by force-torque
 *     Dq  : quaternion delta as integrated by IMU
 *
 * In VectorComposite classes representing deltas, these blocks are identified with the following keys
 *      'P': position delta as integrated by IMU
 *      'V': velocity delta as integrated by IMU
 *      'p': position delta as integrated by FT
 *      'v': velocity delta as integrated by FT
 *      'L': angular momentum delta as integrated by FT
 *      'O': orientation delta as integrated by IMU
 *
 * When Deltas are represented in vector form VectorXd, the order of the blocks is "PVpvLO".
 *
 * In Frame classes, blocks are identified with the following keys
 *      'P': position
 *      'V': velocity
 *      'L': angular momentum
 *      'O': orientation
 *
 * When states are represented in vector form VectorXd, the order of the blocks is "PVLO".
 * Such Frame states are therefore of size 13, VectorXd(13).
 *
 * Functions involving group-like operations on deltas are listed below:
 *
 *   - identity: I = Delta at the origin,
 *      with Dpi = [0,0,0], Dvi = [0,0,0], Dpd = [0,0,0], Dvd = [0,0,0], DL = [0,0,0], Dq = [0,0,0,1]
 *   - inverse: so that D * D.inv = I
 *   - compose: Dc = D1 * D2
 *   - between: Db = D1.inv * D2, so that D2 = D1 * Db
 *   - composeOverState: x2 = x1 [*] D'
 *      where D' is a Delta with just the IMU or the FT parts for the "PV" blocks,
 *      that is "PVLO" or "pvLO",
 *      as decided by the user when providing the delta.
 *   - composeOverDelta: x2 = x1 [*] D
 *      where x is a state "PVLO" and D is a delta "PVpvLO".
 *      A second parameter `fti::Method` is provided
 *      to select between IMU "PV" and FT "pv" delta blocks to create the state "PVLO".
 *   - betweenStates: D = x2 (-) x1, so that x2 = x1 [*] D
 *      where D is a full delta "PVpvLO", but the "PV" and "pv" blocks are equal.
 *   - log_fti: go from Delta manifold to tangent space (equivalent to log() in rotations)
 *   - exp_fti: go from tangent space to delta manifold (equivalent to exp() in rotations)
 *   - plus: D2 = D1 * exp_fti(d)
 *   - diff: d = log_fti( D2 (-) D1 )
 *
 * For creating the deltas from real data, the following vectors are defined:
 *   - data[6+np] : [acc[3], gyro[3], prop_speeds[np]], with acc, gyro from IMU, prop_speeds of any size `np`
 *   - bias[12]   : [acc_b[3], gyro_b[3], force_b[3], torque_b[3]] each a 3d vector
 *   - model[7]   : [com[3], inertia[3], mass[1]], with
 *          - com[3]    : position of CoM wrt. Base frame
 *          - inertia[3]: a 3-vector with [Ixx, Iyy, Izz], so that the inertia matrix is inertia.asDiagonal()
 *          - mass[1]   : a scalar, in Kg
 *   - tangent[12]: [acc, angvel, force, torque], a vector of intermediate magnitudes, each piece a 3-vector
 *
 * The design of the math in this file assumes the IMU frame matches the Base frame exactly.
 *
 * Also a structure Propeller holds important data for a propeller:
 *   - position : position wrt Base frame
 *   - rotation : rotation matrix wrt. Base frame
 *   - c_T      : thrust param, thrust = c_T * n^2, with `n` propeller speed in rad/s
 *   - c_M      : torque param, torque = - dir * c_M * n^2, with `n` propeller speed in rad/s, and `dir` the blade
 * direction below
 *   - dir_ccw  : direction of blades: clockwise = -1, counter-clockwise = -1.
 *
 * The following functions are defined, with all Jacobians:
 *   - data2tangent  : obtain tangent from data
 *   - forces2acc    : obtain acceleration from forces
 *   - tangent2delta : obtain delta from tangent
 *   - data2delta    : obtain deltas from data (uses the three above)
 */

namespace wolf
{
namespace bodydynamics
{
namespace fti
{
using namespace Eigen;

/** \brief method to compose deltas with states
 */
enum Method
{
    IMU,  ///< Will take IMU parts of the Delta for P and V
    FT    ///< Will take force-torque parts of the delta for P and V
};

struct Propeller
{
    Propeller(const Vector3d& _position,
              const Matrix3d& _rotation,
              const double&   _direction_ccw,  /// 1: Counter clock wise (CCW) seen from above; -1: CW
              double          _c_T,
              double          _c_M)
        : position(_position), rotation(_rotation), direction_ccw(_direction_ccw), c_T(_c_T), c_M(_c_M){};

    Vector3d position;
    Matrix3d rotation;
    double   direction_ccw;
    double   c_T;  // thrust
    double   c_M;  // torque
};

template <typename D1, typename D2, typename D3, typename D4, typename D5, typename D6>
inline void identity(MatrixBase<D1>&     dpi,
                     MatrixBase<D2>&     dvi,
                     MatrixBase<D3>&     dpd,
                     MatrixBase<D4>&     dvd,
                     MatrixBase<D5>&     dL,
                     QuaternionBase<D6>& dq)
{
    dpi = MatrixBase<D1>::Zero(3, 1);
    dvi = MatrixBase<D2>::Zero(3, 1);
    dpd = MatrixBase<D3>::Zero(3, 1);
    dvd = MatrixBase<D4>::Zero(3, 1);
    dL  = MatrixBase<D5>::Zero(3, 1);
    dq  = QuaternionBase<D6>::Identity();
}

template <typename D1, typename D2, typename D3, typename D4, typename D5, typename D6>
inline void identity(MatrixBase<D1>& dpi,
                     MatrixBase<D2>& dvi,
                     MatrixBase<D3>& dpd,
                     MatrixBase<D4>& dvd,
                     MatrixBase<D5>& dL,
                     MatrixBase<D6>& dq)
{
    typedef typename D1::Scalar T1;
    typedef typename D2::Scalar T2;
    typedef typename D3::Scalar T3;
    typedef typename D4::Scalar T4;
    typedef typename D5::Scalar T5;
    typedef typename D6::Scalar T6;
    dpi << T1(0), T1(0), T1(0);
    dvi << T2(0), T2(0), T2(0);
    dpd << T3(0), T3(0), T3(0);
    dvd << T4(0), T4(0), T4(0);
    dL << T5(0), T5(0), T5(0);
    dq << T6(0), T6(0), T6(0), T6(1);
}

template <typename T = double>
inline Matrix<T, 19, 1> identity()
{
    Matrix<T, 19, 1> ret;
    ret << T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0), T(0),
        T(1);
    return ret;
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12,
          class T>
inline void inverse(MatrixBase<D1>&      dpi,
                    MatrixBase<D2>&      dvi,
                    MatrixBase<D3>&      dpd,
                    MatrixBase<D4>&      dvd,
                    MatrixBase<D5>&      dL,
                    QuaternionBase<D6>&  dq,
                    const T              dt,
                    MatrixBase<D7>&      idpi,
                    MatrixBase<D8>&      idvi,
                    MatrixBase<D9>&      idpd,
                    MatrixBase<D10>&     idvd,
                    MatrixBase<D11>&     idL,
                    QuaternionBase<D12>& idq)
{
    MatrixSizeCheck<3, 1>::check(dpi);
    MatrixSizeCheck<3, 1>::check(dvi);
    MatrixSizeCheck<3, 1>::check(dpd);
    MatrixSizeCheck<3, 1>::check(dvd);
    MatrixSizeCheck<3, 1>::check(dL);
    MatrixSizeCheck<3, 1>::check(idpi);
    MatrixSizeCheck<3, 1>::check(idvi);
    MatrixSizeCheck<3, 1>::check(idpd);
    MatrixSizeCheck<3, 1>::check(idvd);
    MatrixSizeCheck<3, 1>::check(idL);

    idpi = -(dq.conjugate() * (dpi - dvi * typename D3::Scalar(dt)));
    idvi = -(dq.conjugate() * dvi);
    idpd = -(dq.conjugate() * (dpd - dvd * typename D3::Scalar(dt)));
    idvd = -(dq.conjugate() * dvd);
    idL  = -(dq.conjugate() * dL);
    idq  = (dq.conjugate());
}

template <typename D1, typename D2, class T>
inline void inverse(const MatrixBase<D1>& d, T dt, MatrixBase<D2>& id)
{
    MatrixSizeCheck<19, 1>::check(d);
    MatrixSizeCheck<19, 1>::check(id);

    Map<const Matrix<typename D1::Scalar, 3, 1>> dpi(&d(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvi(&d(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dpd(&d(6));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvd(&d(9));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dL(&d(12));
    Map<const Quaternion<typename D1::Scalar>>   dq(&d(15));
    Map<Matrix<typename D2::Scalar, 3, 1>>       idpi(&id(0));
    Map<Matrix<typename D2::Scalar, 3, 1>>       idvi(&id(3));
    Map<Matrix<typename D2::Scalar, 3, 1>>       idpd(&id(6));
    Map<Matrix<typename D2::Scalar, 3, 1>>       idvd(&id(9));
    Map<Matrix<typename D2::Scalar, 3, 1>>       idL(&id(12));
    Map<Quaternion<typename D1::Scalar>>         idq(&id(15));

    inverse(dpi, dvi, dpd, dvd, dL, dq, dt, idpi, idvi, idpd, idvd, idL, idq);
}

template <typename D, class T>
inline Matrix<typename D::Scalar, 19, 1> inverse(const MatrixBase<D>& d, T dt)
{
    Matrix<typename D::Scalar, 19, 1> id;
    inverse(d, dt, id);
    return id;
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12,
          typename D19,
          typename D14,
          typename D15,
          typename D16,
          typename D17,
          typename D18,
          class T>
inline void compose(const MatrixBase<D1>&      dpi1,
                    const MatrixBase<D2>&      dvi1,
                    const MatrixBase<D3>&      dpd1,
                    const MatrixBase<D4>&      dvd1,
                    const MatrixBase<D5>&      dL1,
                    const QuaternionBase<D6>&  dq1,
                    const MatrixBase<D7>&      dpi2,
                    const MatrixBase<D8>&      dvi2,
                    const MatrixBase<D9>&      dpd2,
                    const MatrixBase<D10>&     dvd2,
                    const MatrixBase<D11>&     dL2,
                    const QuaternionBase<D12>& dq2,
                    const T                    dt,
                    MatrixBase<D19>&           sum_pi,
                    MatrixBase<D14>&           sum_vi,
                    MatrixBase<D15>&           sum_pd,
                    MatrixBase<D16>&           sum_vd,
                    MatrixBase<D17>&           sum_L,
                    QuaternionBase<D18>&       sum_q)
{
    MatrixSizeCheck<3, 1>::check(dpi1);
    MatrixSizeCheck<3, 1>::check(dvi1);
    MatrixSizeCheck<3, 1>::check(dpd1);
    MatrixSizeCheck<3, 1>::check(dvd1);
    MatrixSizeCheck<3, 1>::check(dL1);
    MatrixSizeCheck<3, 1>::check(dpi2);
    MatrixSizeCheck<3, 1>::check(dvi2);
    MatrixSizeCheck<3, 1>::check(dpd2);
    MatrixSizeCheck<3, 1>::check(dvd2);
    MatrixSizeCheck<3, 1>::check(dL2);
    MatrixSizeCheck<3, 1>::check(sum_pi);
    MatrixSizeCheck<3, 1>::check(sum_vi);
    MatrixSizeCheck<3, 1>::check(sum_pd);
    MatrixSizeCheck<3, 1>::check(sum_vd);
    MatrixSizeCheck<3, 1>::check(sum_L);

    sum_pi = dpi1 + dvi1 * dt + dq1 * dpi2;
    sum_vi = dvi1 + dq1 * dvi2;
    sum_pd = dpd1 + dvd1 * dt + dq1 * dpd2;
    sum_vd = dvd1 + dq1 * dvd2;
    sum_L  = dL1 + dq1 * dL2;
    sum_q  = dq1 * dq2;
}

template <typename D1, typename D2, typename D3, class T>
inline void compose(const MatrixBase<D1>& d1, const MatrixBase<D2>& d2, T dt, MatrixBase<D3>& sum)
{
    MatrixSizeCheck<19, 1>::check(d1);
    MatrixSizeCheck<19, 1>::check(d2);
    MatrixSizeCheck<19, 1>::check(sum);

    Map<const Matrix<typename D1::Scalar, 3, 1>> dpi1(&d1(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvi1(&d1(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dpd1(&d1(6));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvd1(&d1(9));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dL1(&d1(12));
    Map<const Quaternion<typename D1::Scalar>>   dq1(&d1(15));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpi2(&d2(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvi2(&d2(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpd2(&d2(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvd2(&d2(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dL2(&d2(12));
    Map<const Quaternion<typename D2::Scalar>>   dq2(&d2(15));
    Map<Matrix<typename D3::Scalar, 3, 1>>       sum_pi(&sum(0));
    Map<Matrix<typename D3::Scalar, 3, 1>>       sum_vi(&sum(3));
    Map<Matrix<typename D3::Scalar, 3, 1>>       sum_pd(&sum(6));
    Map<Matrix<typename D3::Scalar, 3, 1>>       sum_vd(&sum(9));
    Map<Matrix<typename D3::Scalar, 3, 1>>       sum_L(&sum(12));
    Map<Quaternion<typename D3::Scalar>>         sum_q(&sum(15));

    compose(dpi1, dvi1, dpd1, dvd1, dL1, dq1, dpi2, dvi2, dpd2, dvd2, dL2, dq2, dt, sum_pi, sum_vi, sum_pd, sum_vd,
            sum_L, sum_q);
}

template <typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 19, 1> compose(const MatrixBase<D1>& d1, const MatrixBase<D2>& d2, T dt)
{
    Matrix<typename D1::Scalar, 19, 1> ret;
    compose(d1, d2, dt, ret);
    return ret;
}

template <typename D1, typename D2, typename D3, typename D4, typename D5, class T>
inline void compose(const MatrixBase<D1>& d1,
                    const MatrixBase<D2>& d2,
                    T                     dt,
                    MatrixBase<D3>&       sum,
                    MatrixBase<D4>&       J_sum_d1,
                    MatrixBase<D5>&       J_sum_d2)
{
    MatrixSizeCheck<19, 1>::check(d1);
    MatrixSizeCheck<19, 1>::check(d2);
    MatrixSizeCheck<19, 1>::check(sum);
    MatrixSizeCheck<18, 18>::check(J_sum_d1);
    MatrixSizeCheck<18, 18>::check(J_sum_d2);

    // Some useful temporaries
    Map<const Quaternion<typename D1::Scalar>>   dq1(&d1(15));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpi2(&d2(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvi2(&d2(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpd2(&d2(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvd2(&d2(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dL2(&d2(12));
    Map<const Quaternion<typename D2::Scalar>>   dq2(&d2(15));

    Matrix<typename D1::Scalar, 3, 3> dR1 = dq1.matrix();
    Matrix<typename D2::Scalar, 3, 3> dR2 = dq2.matrix();

    // JACOBIANS OF COMPOSE()                             JAC WRT D1                          JAC WRT D2
    //                                       dpi1 dvi1  dpd1  dvd1  dL1     dR1       dpi2 dvi2 dpd2 dvd2 dL2 dR2
    //                                     ------------------------------------------  --------------------------
    // sum_pi = dpi1  + dvi1*dt + dq1*dpi2; //  I,  I*dt, 0,  0   , 0, -dR1 * dpi2_x  // dR1, 0,  0,  0,  0,  0
    // sum_vi = dvi1  +           dq1*dvi2; //  0,  I   , 0,  0   , 0, -dR1 * dvi2_x  //  0, dR1, 0,  0,  0,  0
    // sum_pd = dpd1  + dvd1*dt + dq1*dpd2; //  0,  0   , I,  I*dt, 0, -dR1 * dpd2_x  //  0,  0, dR1, 0,  0,  0
    // sum_vd = dvd1  +           dq1*dvd2; //  0,  0   , 0,  I   , 0, -dR1 * dvd2_x  //  0,  0,  0, dR1, 0,  0
    // sum_L  = dL1   +           dq1*dL2;  //  0,  0   , 0,  0   , I, -dR1 * dL2_x   //  0,  0,  0,  0, dR1, 0 
    // sum_q  =                   dq1*dq2;  //  0,  0   , 0,  0   , 0,  dR2.tr        //  0,  0,  0,  0,  0,  I

    // // Jac wrt first delta
    J_sum_d1.setIdentity();
    J_sum_d1.template block<3, 3>(0, 3)   = Matrix3d::Identity() * dt;
    J_sum_d1.template block<3, 3>(0, 15)  = -dR1 * skew(dpi2);
    J_sum_d1.template block<3, 3>(3, 15)  = -dR1 * skew(dvi2);
    J_sum_d1.template block<3, 3>(6, 9)   = Matrix3d::Identity() * dt;
    J_sum_d1.template block<3, 3>(6, 15)  = -dR1 * skew(dpd2);
    J_sum_d1.template block<3, 3>(9, 15)  = -dR1 * skew(dvd2);
    J_sum_d1.template block<3, 3>(12, 15) = -dR1 * skew(dL2);
    J_sum_d1.template block<3, 3>(15, 15) = dR2.transpose();

    // // // Jac wrt second delta
    J_sum_d2.setIdentity();
    J_sum_d2.template block<3, 3>(0, 0)   = dR1;
    J_sum_d2.template block<3, 3>(3, 3)   = dR1;
    J_sum_d2.template block<3, 3>(6, 6)   = dR1;
    J_sum_d2.template block<3, 3>(9, 9)   = dR1;
    J_sum_d2.template block<3, 3>(12, 12) = dR1;

    // compose deltas -- done here to avoid aliasing when calling with input
    // `d1` and result `sum` referencing the same variable
    compose(d1, d2, dt, sum);
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12,
          typename D13,
          typename D14,
          typename D15,
          typename D16,
          typename D17,
          typename D18,
          class T>
inline void between(const MatrixBase<D1>&      dpi1,
                    const MatrixBase<D2>&      dvi1,
                    const MatrixBase<D3>&      dpd1,
                    const MatrixBase<D4>&      dvd1,
                    const MatrixBase<D5>&      dL1,
                    const QuaternionBase<D6>&  dq1,
                    const MatrixBase<D7>&      dpi2,
                    const MatrixBase<D8>&      dvi2,
                    const MatrixBase<D9>&      dpd2,
                    const MatrixBase<D10>&     dvd2,
                    const MatrixBase<D11>&     dL2,
                    const QuaternionBase<D12>& dq2,
                    const T                    dt,
                    MatrixBase<D13>&           bet_pi,
                    MatrixBase<D14>&           bet_vi,
                    MatrixBase<D15>&           bet_pd,
                    MatrixBase<D16>&           bet_vd,
                    MatrixBase<D17>&           bet_L,
                    QuaternionBase<D18>&       bet_q)
{
    MatrixSizeCheck<3, 1>::check(dpi1);
    MatrixSizeCheck<3, 1>::check(dvi1);
    MatrixSizeCheck<3, 1>::check(dpd1);
    MatrixSizeCheck<3, 1>::check(dvd1);
    MatrixSizeCheck<3, 1>::check(dL1);
    MatrixSizeCheck<3, 1>::check(dpi2);
    MatrixSizeCheck<3, 1>::check(dvi2);
    MatrixSizeCheck<3, 1>::check(dpd2);
    MatrixSizeCheck<3, 1>::check(dvd2);
    MatrixSizeCheck<3, 1>::check(dL2);
    MatrixSizeCheck<3, 1>::check(bet_pi);
    MatrixSizeCheck<3, 1>::check(bet_vi);
    MatrixSizeCheck<3, 1>::check(bet_pd);
    MatrixSizeCheck<3, 1>::check(bet_vd);
    MatrixSizeCheck<3, 1>::check(bet_L);

    bet_pi = dq1.conjugate() * (dpi2 - dpi1 - dvi1 * dt);
    bet_vi = dq1.conjugate() * (dvi2 - dvi1);
    bet_pd = dq1.conjugate() * (dpd2 - dpd1 - dvd1 * dt);
    bet_vd = dq1.conjugate() * (dvd2 - dvd1);
    bet_L  = dq1.conjugate() * (dL2 - dL1);
    bet_q  = dq1.conjugate() * dq2;
}

template <typename D1, typename D2, typename D3, class T>
inline void between(const MatrixBase<D1>& d1, const MatrixBase<D2>& d2, T dt, MatrixBase<D3>& d2_minus_d1)
{
    MatrixSizeCheck<19, 1>::check(d1);
    MatrixSizeCheck<19, 1>::check(d2);
    MatrixSizeCheck<19, 1>::check(d2_minus_d1);

    Map<const Matrix<typename D1::Scalar, 3, 1>> dpi1(&d1(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvi1(&d1(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dpd1(&d1(6));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvd1(&d1(9));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dL1(&d1(12));
    Map<const Quaternion<typename D1::Scalar>>   dq1(&d1(15));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpi2(&d2(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvi2(&d2(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpd2(&d2(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvd2(&d2(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dL2(&d2(12));
    Map<const Quaternion<typename D2::Scalar>>   dq2(&d2(15));
    Map<Matrix<typename D3::Scalar, 3, 1>>       bet_pi(&d2_minus_d1(0));
    Map<Matrix<typename D3::Scalar, 3, 1>>       bet_vi(&d2_minus_d1(3));
    Map<Matrix<typename D3::Scalar, 3, 1>>       bet_pd(&d2_minus_d1(6));
    Map<Matrix<typename D3::Scalar, 3, 1>>       bet_vd(&d2_minus_d1(9));
    Map<Matrix<typename D3::Scalar, 3, 1>>       bet_L(&d2_minus_d1(12));
    Map<Quaternion<typename D3::Scalar>>         bet_q(&d2_minus_d1(15));

    between(dpi1, dvi1, dpd1, dvd1, dL1, dq1, dpi2, dvi2, dpd2, dvd2, dL2, dq2, dt, bet_pi, bet_vi, bet_pd, bet_vd,
            bet_L, bet_q);
}

template <typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 19, 1> between(const MatrixBase<D1>& d1, const MatrixBase<D2>& d2, T dt)
{
    Matrix<typename D1::Scalar, 19, 1> diff;
    between(d1, d2, dt, diff);
    return diff;
}

inline VectorComposite composeOverState(const VectorComposite& x,
                                        const VectorComposite& d,
                                        double                 dt,
                                        Method                 method = IMU)
{
    // Maps to have nice names
    Map<const Vector3d>    p(x.at('P').data());  // state POVL
    Map<const Vector3d>    v(x.at('V').data());
    Map<const Quaterniond> q(x.at('O').data());
    Map<const Vector3d>    L(x.at('L').data());
    Map<const Vector3d>    dpi(d.at('P').data());  // delta PVO from IMU
    Map<const Vector3d>    dvi(d.at('V').data());
    Map<const Quaterniond> dq(d.at('O').data());
    Map<const Vector3d>    dpd(d.at('p').data());  // delta pvL from FT
    Map<const Vector3d>    dvd(d.at('v').data());
    Map<const Vector3d>    dL(d.at('L').data());

    const Vector3d& g_dt = gravity() * dt;

    VectorComposite x_plus_d;
    switch (method)
    {
        case IMU:
        default:
            x_plus_d['P'] = p + v * dt + 0.5 * g_dt * dt + q * dpi;
            x_plus_d['V'] = v + g_dt + q * dvi;
            break;
        case FT:
            x_plus_d['P'] = p + v * dt + 0.5 * g_dt * dt + q * dpd;
            x_plus_d['V'] = v + g_dt + q * dvd;
            break;
    }
    x_plus_d['L'] = L + q * dL;
    x_plus_d['O'] = (q * dq).coeffs();

    return x_plus_d;
}

template <typename D1, typename D2, typename D3, class T>
inline void composeOverState(const MatrixBase<D1>& x,
                             const MatrixBase<D2>& d,
                             T                     dt,
                             MatrixBase<D3>&       x_plus_d,
                             Method                method = IMU)
{
    MatrixSizeCheck<13, 1>::check(x);
    MatrixSizeCheck<19, 1>::check(d);
    MatrixSizeCheck<13, 1>::check(x_plus_d);

    Map<const Matrix<typename D1::Scalar, 3, 1>> p(&x(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> v(&x(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> L(&x(6));
    Map<const Quaternion<typename D1::Scalar>>   q(&x(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpi(&d(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvi(&d(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpd(&d(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvd(&d(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dL(&d(12));
    Map<const Quaternion<typename D2::Scalar>>   dq(&d(15));
    Map<Matrix<typename D3::Scalar, 3, 1>>       p_plus_d(&x_plus_d(0));
    Map<Matrix<typename D3::Scalar, 3, 1>>       v_plus_d(&x_plus_d(3));
    Map<Matrix<typename D3::Scalar, 3, 1>>       L_plus_d(&x_plus_d(6));
    Map<Quaternion<typename D3::Scalar>>         q_plus_d(&x_plus_d(9));

    const Vector3d& g_dt = gravity() * dt;

    switch (method)
    {
        case IMU:
        default:
            p_plus_d = p + v * dt + 0.5 * g_dt * dt + q * dpi;
            v_plus_d = v + g_dt + q * dvi;
            break;
        case FT:
            p_plus_d = p + v * dt + 0.5 * g_dt * dt + q * dpd;
            v_plus_d = v + g_dt + q * dvd;
            break;
    }
    L_plus_d = L + q * dL;
    q_plus_d = q * dq;
}

template <typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 13, 1> composeOverState(const MatrixBase<D1>& x,
                                                           const MatrixBase<D2>& d,
                                                           T                     dt,
                                                           Method                method = IMU)
{
    Matrix<typename D1::Scalar, 13, 1> ret;
    composeOverState(x, d, dt, ret, method);
    return ret;
}

template <typename D1,
          typename D2,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D11,
          typename D12,
          typename D13,
          typename D14,
          typename D15,
          typename D16,
          typename D17,
          typename D18,
          class T>
inline void betweenStates(const MatrixBase<D1>&      p1,
                          const MatrixBase<D2>&      v1,
                          const MatrixBase<D5>&      L1,
                          const QuaternionBase<D6>&  q1,
                          const MatrixBase<D7>&      p2,
                          const MatrixBase<D8>&      v2,
                          const MatrixBase<D11>&     L2,
                          const QuaternionBase<D12>& q2,
                          const T                    dt,
                          MatrixBase<D13>&           dpi,
                          MatrixBase<D14>&           dvi,
                          MatrixBase<D15>&           dpd,
                          MatrixBase<D16>&           dvd,
                          MatrixBase<D17>&           dL,
                          QuaternionBase<D18>&       dq)
{
    MatrixSizeCheck<3, 1>::check(p1);
    MatrixSizeCheck<3, 1>::check(v1);
    MatrixSizeCheck<3, 1>::check(L1);
    MatrixSizeCheck<3, 1>::check(p2);
    MatrixSizeCheck<3, 1>::check(v2);
    MatrixSizeCheck<3, 1>::check(L2);
    MatrixSizeCheck<3, 1>::check(dpi);
    MatrixSizeCheck<3, 1>::check(dvi);
    MatrixSizeCheck<3, 1>::check(dpd);
    MatrixSizeCheck<3, 1>::check(dvd);
    MatrixSizeCheck<3, 1>::check(dL);

    dpi = q1.conjugate() * (p2 - p1 - v1 * dt - 0.5 * gravity() * dt * dt);
    dvi = q1.conjugate() * (v2 - v1 - gravity() * dt);
    dpd = dpi;
    dvd = dvi;
    dL  = q1.conjugate() * (L2 - L1);
    dq  = q1.conjugate() * q2;
}

template <typename D1, typename D2, typename D3, class T>
inline void betweenStates(const MatrixBase<D1>& x1, const MatrixBase<D2>& x2, T dt, MatrixBase<D3>& x2_minus_x1)
{
    MatrixSizeCheck<13, 1>::check(x1);
    MatrixSizeCheck<13, 1>::check(x2);
    MatrixSizeCheck<19, 1>::check(x2_minus_x1);

    Map<const Matrix<typename D1::Scalar, 3, 1>> p1(&x1(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> v1(&x1(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> L1(&x1(6));
    Map<const Quaternion<typename D1::Scalar>>   q1(&x1(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> p2(&x2(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> v2(&x2(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> L2(&x2(6));
    Map<const Quaternion<typename D2::Scalar>>   q2(&x2(9));
    Map<Matrix<typename D3::Scalar, 3, 1>>       dpi(&x2_minus_x1(0));
    Map<Matrix<typename D3::Scalar, 3, 1>>       dvi(&x2_minus_x1(3));
    Map<Matrix<typename D3::Scalar, 3, 1>>       dpd(&x2_minus_x1(6));
    Map<Matrix<typename D3::Scalar, 3, 1>>       dvd(&x2_minus_x1(9));
    Map<Matrix<typename D3::Scalar, 3, 1>>       dL(&x2_minus_x1(12));
    Map<Quaternion<typename D3::Scalar>>         dq(&x2_minus_x1(15));

    betweenStates(p1, v1, L1, q1, p2, v2, L2, q2, dt, dpi, dvi, dpd, dvd, dL, dq);
}

template <typename D1, typename D2, class T>
inline Matrix<typename D1::Scalar, 19, 1> betweenStates(const MatrixBase<D1>& x1, const MatrixBase<D2>& x2, T dt)
{
    Matrix<typename D1::Scalar, 19, 1> ret;
    betweenStates(x1, x2, dt, ret);
    return ret;
}

inline void betweenStates(const VectorComposite& x1, const VectorComposite& x2, double dt, VectorComposite& ret)
{
    Map<const Quaterniond> q1(x1.at('O').data());
    Map<const Quaterniond> q2(x2.at('O').data());

    ret['P'] = q1.conjugate() * (x2.at('P') - x1.at('P') - x1.at('V') * dt - 0.5 * gravity() * dt * dt);
    ret['V'] = q1.conjugate() * (x2.at('V') - x1.at('V') - gravity() * dt);
    ret['p'] = ret.at('P');
    ret['v'] = ret.at('V');
    ret['L'] = q1.conjugate() * (x2.at('L') - x1.at('L'));
    ret['O'] = (q1.conjugate() * q2).coeffs();
}

inline VectorComposite betweenStates(const VectorComposite& x1, const VectorComposite& x2, double dt)
{
    VectorComposite ret;
    betweenStates(x1, x2, dt, ret);
    return ret;
}

template <typename Derived>
Matrix<typename Derived::Scalar, 18, 1> log_fti(const MatrixBase<Derived>& delta_in)
{
    MatrixSizeCheck<19, 1>::check(delta_in);

    Matrix<typename Derived::Scalar, 18, 1> ret;

    Map<const Matrix<typename Derived::Scalar, 3, 1>> dpi_in(&delta_in(0));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dvi_in(&delta_in(3));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dpd_in(&delta_in(6));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dvd_in(&delta_in(9));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dL_in(&delta_in(12));
    Map<const Quaternion<typename Derived::Scalar>>   dq_in(&delta_in(15));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dpi_ret(&ret(0));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dvi_ret(&ret(3));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dpd_ret(&ret(6));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dvd_ret(&ret(9));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dL_ret(&ret(12));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       do_ret(&ret(15));

    dpi_ret = dpi_in;
    dvi_ret = dvi_in;
    dpd_ret = dpd_in;
    dvd_ret = dvd_in;
    dL_ret  = dL_in;
    do_ret  = log_q(dq_in);

    return ret;
}

template <typename Derived>
Matrix<typename Derived::Scalar, 19, 1> exp_fti(const MatrixBase<Derived>& d_alg_in)
{
    MatrixSizeCheck<18, 1>::check(d_alg_in);

    Matrix<typename Derived::Scalar, 19, 1> ret;

    Map<const Matrix<typename Derived::Scalar, 3, 1>> dpi_alg_in(&d_alg_in(0));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dvi_alg_in(&d_alg_in(3));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dpd_alg_in(&d_alg_in(6));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dvd_alg_in(&d_alg_in(9));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> dL_alg_in(&d_alg_in(12));
    Map<const Matrix<typename Derived::Scalar, 3, 1>> do_alg_in(&d_alg_in(15));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dpi_ret(&ret(0));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dvi_ret(&ret(3));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dpd_ret(&ret(6));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dvd_ret(&ret(9));
    Map<Matrix<typename Derived::Scalar, 3, 1>>       dL_ret(&ret(12));
    Map<Quaternion<typename Derived::Scalar>>         dq_ret(&ret(15));

    dpi_ret = dpi_alg_in;
    dvi_ret = dvi_alg_in;
    dpd_ret = dpd_alg_in;
    dvd_ret = dvd_alg_in;
    dL_ret  = dL_alg_in;
    dq_ret  = exp_q(do_alg_in);

    return ret;
}

template <typename D1, typename D2, typename D3>
void exp_fti(const MatrixBase<D1>& _tangent, MatrixBase<D2>& _delta, MatrixBase<D3>& _jacobian_delta_tangent)
{
    MatrixSizeCheck<18, 1>::check(_tangent);
    MatrixSizeCheck<19, 1>::check(_delta);
    MatrixSizeCheck<18, 18>::check(_jacobian_delta_tangent);

    _delta = fti::exp_fti(_tangent);
    _jacobian_delta_tangent.setIdentity();
    _jacobian_delta_tangent.template block<3, 3>(15, 15) = jac_SO3_right(_tangent.template segment<3>(15));
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12,
          typename D13,
          typename D14,
          typename D15,
          typename D16,
          typename D17,
          typename D18>
inline void plus(const MatrixBase<D1>&     dpi1,
                 const MatrixBase<D2>&     dvi1,
                 const MatrixBase<D3>&     dpd1,
                 const MatrixBase<D4>&     dvd1,
                 const MatrixBase<D5>&     dL1,
                 const QuaternionBase<D6>& dq1,
                 const MatrixBase<D7>&     diff_pi2,
                 const MatrixBase<D8>&     diff_vi2,
                 const MatrixBase<D9>&     diff_pd2,
                 const MatrixBase<D10>&    diff_vd2,
                 const MatrixBase<D11>&    diff_L2,
                 const MatrixBase<D12>&    diff_o2,
                 MatrixBase<D13>&          plus_pi,
                 MatrixBase<D14>&          plus_vi,
                 MatrixBase<D15>&          plus_pd,
                 MatrixBase<D16>&          plus_vd,
                 MatrixBase<D17>&          plus_L,
                 QuaternionBase<D18>&      plus_q)
{
    plus_pi = dpi1 + diff_pi2;
    plus_vi = dvi1 + diff_vi2;
    plus_pd = dpd1 + diff_pd2;
    plus_vd = dvd1 + diff_vd2;
    plus_L  = dL1 + diff_L2;
    plus_q  = dq1 * exp_q(diff_o2);
}

// Composite plus
template <typename D1, typename D2, typename D3>
inline void plus(const MatrixBase<D1>& d1, const MatrixBase<D2>& diff2, MatrixBase<D3>& d)
{
    Map<const Matrix<typename D1::Scalar, 3, 1>> dpi1(&d1(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvi1(&d1(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dpd1(&d1(6));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvd1(&d1(9));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dL1(&d1(12));
    Map<const Quaternion<typename D1::Scalar>>   dq1(&d1(15));
    Map<const Matrix<typename D2::Scalar, 3, 1>> diff_dpi2(&diff2(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> diff_dvi2(&diff2(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> diff_dpd2(&diff2(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>> diff_dvd2(&diff2(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> diff_dL2(&diff2(12));
    Map<const Matrix<typename D2::Scalar, 3, 1>> diff_do2(&diff2(15));
    Map<Matrix<typename D3::Scalar, 3, 1>>       d_pi(&d(0));
    Map<Matrix<typename D3::Scalar, 3, 1>>       d_vi(&d(3));
    Map<Matrix<typename D3::Scalar, 3, 1>>       d_pd(&d(6));
    Map<Matrix<typename D3::Scalar, 3, 1>>       d_vd(&d(9));
    Map<Matrix<typename D3::Scalar, 3, 1>>       d_L(&d(12));
    Map<Quaternion<typename D3::Scalar>>         d_q(&d(15));

    plus(dpi1, dvi1, dpd1, dvd1, dL1, dq1, diff_dpi2, diff_dvi2, diff_dpd2, diff_dvd2, diff_dL2, diff_do2, d_pi, d_vi,
         d_pd, d_vd, d_L, d_q);
}

template <typename D1, typename D2>
inline Matrix<typename D1::Scalar, 19, 1> plus(const MatrixBase<D1>& d1, const MatrixBase<D2>& d2)
{
    Matrix<typename D1::Scalar, 19, 1> ret;
    plus(d1, d2, ret);
    return ret;
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12,
          typename D13,
          typename D14,
          typename D15,
          typename D16,
          typename D17,
          typename D18>
inline void diff(const MatrixBase<D1>&      dpi1,
                 const MatrixBase<D2>&      dvi1,
                 const MatrixBase<D3>&      dpd1,
                 const MatrixBase<D4>&      dvd1,
                 const MatrixBase<D5>&      dL1,
                 const QuaternionBase<D6>&  dq1,
                 const MatrixBase<D7>&      dpi2,
                 const MatrixBase<D8>&      dvi2,
                 const MatrixBase<D9>&      dpd2,
                 const MatrixBase<D10>&     dvd2,
                 const MatrixBase<D11>&     dL2,
                 const QuaternionBase<D12>& dq2,
                 MatrixBase<D13>&           diff_pi,
                 MatrixBase<D14>&           diff_vi,
                 MatrixBase<D15>&           diff_pd,
                 MatrixBase<D16>&           diff_vd,
                 MatrixBase<D17>&           diff_L,
                 MatrixBase<D18>&           diff_o)
{
    diff_pi = dpi2 - dpi1;
    diff_vi = dvi2 - dvi1;
    diff_pd = dpd2 - dpd1;
    diff_vd = dvd2 - dvd1;
    diff_L  = dL2 - dL1;
    diff_o  = log_q(dq1.conjugate() * dq2);
}

template <typename D1, typename D2, typename D3>
inline void diff(const MatrixBase<D1>& d1, const MatrixBase<D2>& d2, MatrixBase<D3>& diff_d1_d2)
{
    Map<const Matrix<typename D1::Scalar, 3, 1>> dpi1(&d1(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvi1(&d1(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dpd1(&d1(6));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dvd1(&d1(9));
    Map<const Matrix<typename D1::Scalar, 3, 1>> dL1(&d1(12));
    Map<const Quaternion<typename D1::Scalar>>   dq1(&d1(15));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpi2(&d2(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvi2(&d2(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dpd2(&d2(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dvd2(&d2(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> dL2(&d2(12));
    Map<const Quaternion<typename D1::Scalar>>   dq2(&d2(15));
    Map<Matrix<typename D3::Scalar, 3, 1>>       diff_pi(&diff_d1_d2(0));
    Map<Matrix<typename D3::Scalar, 3, 1>>       diff_vi(&diff_d1_d2(3));
    Map<Matrix<typename D3::Scalar, 3, 1>>       diff_pd(&diff_d1_d2(6));
    Map<Matrix<typename D3::Scalar, 3, 1>>       diff_vd(&diff_d1_d2(9));
    Map<Matrix<typename D3::Scalar, 3, 1>>       diff_L(&diff_d1_d2(12));
    Map<Matrix<typename D3::Scalar, 3, 1>>       diff_o(&diff_d1_d2(15));

    diff(dpi1, dvi1, dpd1, dvd1, dL1, dq1, dpi2, dvi2, dpd2, dvd2, dL2, dq2, diff_pi, diff_vi, diff_pd, diff_vd,
         diff_L, diff_o);
}

template <typename D1, typename D2>
inline Matrix<typename D1::Scalar, 18, 1> diff(const MatrixBase<D1>& d1, const MatrixBase<D2>& d2)
{
    Matrix<typename D1::Scalar, 18, 1> ret;
    diff(d1, d2, ret);
    return ret;
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12>
void data2tangent(const MatrixBase<D1>&       _data_acc,
                  const MatrixBase<D2>&       _data_angvel,
                  const MatrixBase<D3>&       _data_motor_speeds,
                  const MatrixBase<D4>&       _bias_acc,
                  const MatrixBase<D5>&       _bias_angvel,
                  const MatrixBase<D6>&       _bias_force,
                  const MatrixBase<D7>&       _bias_torque,
                  const MatrixBase<D8>&       _com,
                  const std::list<Propeller>& _propellers,
                  MatrixBase<D9>&             _tangent_acc,
                  MatrixBase<D10>&            _tangent_angvel,
                  MatrixBase<D11>&            _tangent_force,
                  MatrixBase<D12>&            _tangent_torque)
{
    // check sizes
    // TODO

    // helpers
    const auto& r_bc = _com;

    _tangent_force.setZero();
    _tangent_torque.setZero();

    unsigned int i = 0;
    for (const Propeller& propeller : _propellers)
    {
        const Vector3d& r_bi       = propeller.position;
        const Matrix3d& R_bi       = propeller.rotation;
        double          direction  = propeller.direction_ccw;
        const double&   c_T        = propeller.c_T;
        const double&   c_M        = propeller.c_M;
        const double    c_M_by_c_T = c_M / c_T;

        const auto& n_i = _data_motor_speeds(i);

        Vector3d f_bi = R_bi.rightCols(1) * c_T * n_i * n_i;

        _tangent_force += f_bi;
        _tangent_torque += (r_bi - r_bc).cross(f_bi) - direction * c_M_by_c_T * f_bi;

        i++;
    }

    _tangent_acc    = _data_acc - _bias_acc;
    _tangent_angvel = _data_angvel - _bias_angvel;
    _tangent_force += _bias_force;
    _tangent_torque += _bias_torque;
}

template <typename D1, typename D2, typename D3, typename D4>
void data2tangent(const MatrixBase<D1>&       _data,
                  const MatrixBase<D2>&       _bias,
                  const MatrixBase<D3>&       _com,
                  const std::list<Propeller>& _propellers,
                  MatrixBase<D4>&             _tangent)
{
    const int                                          Np = _data.size() - 6;
    Map<const Matrix<typename D1::Scalar, 3, 1>>       am(&_data(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>>       wm(&_data(3));
    Map<const Matrix<typename D1::Scalar, Dynamic, 1>> n(&_data(6), Np);

    Map<const Matrix<typename D2::Scalar, 3, 1>> ab(&_bias(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> wb(&_bias(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>> fb(&_bias(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>> mb(&_bias(9));

    Map<Matrix<typename D4::Scalar, 3, 1>> a(&_tangent(0));
    Map<Matrix<typename D4::Scalar, 3, 1>> w(&_tangent(3));
    Map<Matrix<typename D4::Scalar, 3, 1>> f(&_tangent(6));
    Map<Matrix<typename D4::Scalar, 3, 1>> m(&_tangent(9));

    data2tangent(am, wm, n, ab, wb, fb, mb, _com, _propellers, a, w, f, m);
}

template <typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7>
void data2tangent(const MatrixBase<D1>&       _data,
                  const MatrixBase<D2>&       _bias,
                  const MatrixBase<D3>&       _com,
                  const std::list<Propeller>& _propellers,
                  MatrixBase<D4>&             _tangent,
                  MatrixBase<D5>&             _J_tan_data,
                  MatrixBase<D6>&             _J_tan_bias,
                  MatrixBase<D7>&             _J_tan_com)
{
    const unsigned int                                 np = _propellers.size();
    Map<const Matrix<typename D1::Scalar, 3, 1>>       am(&_data(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>>       wm(&_data(3));
    Map<const Matrix<typename D1::Scalar, Dynamic, 1>> n(&_data(6), np);
    Map<const Matrix<typename D2::Scalar, 3, 1>>       ab(&_bias(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>>       wb(&_bias(3));
    Map<const Matrix<typename D2::Scalar, 3, 1>>       fb(&_bias(6));
    Map<const Matrix<typename D2::Scalar, 3, 1>>       mb(&_bias(9));

    Map<Matrix<typename D4::Scalar, 3, 1>> a(&_tangent(0));
    Map<Matrix<typename D4::Scalar, 3, 1>> w(&_tangent(3));
    Map<Matrix<typename D4::Scalar, 3, 1>> f(&_tangent(6));
    Map<Matrix<typename D4::Scalar, 3, 1>> m(&_tangent(9));

    data2tangent(_data, _bias, _com, _propellers, _tangent);

    // blocks
    const unsigned int A = 0;  // acc
    const unsigned int W = 3;  // gyro
    const unsigned int F = 6;  // force
    const unsigned int M = 9;  // torque
    const unsigned int N = 6;  // motor speeds
    const unsigned int C = 0;  // CoM

    const Matrix3d I_33 = Matrix3d::Identity();

    // jac wrt data
    _J_tan_data.setIdentity();
    //    _J_tan_data.block<3,3>(A,A) = I_33;
    //    _J_tan_data.block<3,3>(W,W) = I_33;

    unsigned int i = 0;
    for (const Propeller& propeller : _propellers)
    {
        const auto& ri  = propeller.position;
        const auto& Ri  = propeller.rotation;
        const auto& dir = propeller.direction_ccw;
        const auto& cTi = propeller.c_T;
        const auto& cMi = propeller.c_M;
        const auto& ni  = n(i);

        Vector3d J_fi_ni = 2 * cTi * ni * Ri.rightCols(1);

        Vector3d J_mi_ni = skew(ri - _com) * J_fi_ni - 2 * dir * cMi * ni * Ri.rightCols(1);

        _J_tan_data.block(F, N + i, 3, 1) = J_fi_ni;
        _J_tan_data.block(M, N + i, 3, 1) = J_mi_ni;

        i++;
    }

    // jac wrt bias
    _J_tan_bias.setIdentity();
    _J_tan_bias.template block<3, 3>(A, A) = -I_33;
    _J_tan_bias.template block<3, 3>(W, W) = -I_33;
    //    _J_tan_bias.block<3,3>(F,F) =   I_33;
    //    _J_tan_bias.block<3,3>(M,M) =   I_33;

    // jac wrt com
    _J_tan_com.setZero();
    //    __J_tan_com.block<3,3>(A,C) = 0;
    //    __J_tan_com.block<3,3>(W,C) = 0;
    //    __J_tan_com.block<3,3>(F,C) = 0;

    // The total torque is
    //    m = sum_i m_i
    // with
    //    m_i = (r_i - r_com) x f_bi + (terms not related to r_com).
    // Then,
    //    J_mi_com = [f_bi]x
    // and thus
    //    J_m_com = sum_i [f_bi]x .
    // But since for any set of a_i
    //    sum [a_i]x = [sum a_i]x
    // we have that
    //    J_m_com = sum [f_bi]x = [sum fbi]x = [ft]x
    // where
    //    ft = sum_i f_bi
    // is the total force. To get it, remove the bias which was added in the function:
    //    ft = f - fb

    _J_tan_com.template block<3, 3>(M, C) = skew(f - fb);
}

template <typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, class T>
void forces2acc(const MatrixBase<D1>& _force,
                const MatrixBase<D2>& _torque,
                const MatrixBase<D3>& _angvel,
                const MatrixBase<D4>& _com,
                const MatrixBase<D5>& _inertia,
                const T               _mass,
                MatrixBase<D6>&       _acc)
{
    typedef typename D6::Scalar TT;  // take scalar from return type
    typedef Matrix<TT, 3, 3>    Matrix3t;
    typedef Matrix<TT, 3, 1>    Vector3t;

    const Matrix3t& I     = _inertia.asDiagonal();
    const Matrix3t& I_inv = ((typename D5::Scalar)(1.0) / _inertia.array()).matrix().asDiagonal();

    const Vector3t& angacc = I_inv * (_torque - _angvel.cross(I * _angvel));

    _acc = _force / _mass - angacc.cross(_com) - _angvel.cross(_angvel.cross(_com));
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12,
          class T>
void forces2acc(const MatrixBase<D1>& _force,
                const MatrixBase<D2>& _torque,
                const MatrixBase<D3>& _angvel,
                const MatrixBase<D4>& _com,
                const MatrixBase<D5>& _inertia,
                const T               _mass,
                MatrixBase<D6>&       _acc,
                MatrixBase<D7>&       _J_acc_force,
                MatrixBase<D8>&       _J_acc_torque,
                MatrixBase<D9>&       _J_acc_angvel,
                MatrixBase<D10>&      _J_acc_com,
                MatrixBase<D11>&      _J_acc_inertia,
                MatrixBase<D12>&      _J_acc_mass)
{
    typedef typename D6::Scalar S;  // take scalar from return type
    typedef Matrix<S, 3, 3>     Matrix3t;
    typedef Matrix<S, 3, 1>     Vector3t;

    // constants
    const Matrix3t& I     = _inertia.asDiagonal();
    const Matrix3t& I_inv = ((typename D5::Scalar)(1.0) / _inertia.array()).matrix().asDiagonal();

    // variables
    const Vector3t& angacc   = I_inv * (_torque - _angvel.cross(I * _angvel));  // (36)
    const Matrix3t& angvel_x = skew(_angvel);
    const Matrix3t& com_x    = skew(_com);
    const Matrix3t& angacc_x = skew(angacc);

    _acc = _force / _mass - angacc.cross(_com) - _angvel.cross(_angvel.cross(_com));  // (35)

    // Jacobian blocks
    const Matrix3t& tmp                = (_angvel.cross(I * _angvel) - _torque).asDiagonal();
    const Matrix3t& J_Iw_i             = _angvel.asDiagonal();                        // (58)
    const Matrix3t& J_angacc_inertia   = I_inv * (I_inv * tmp - angvel_x * J_Iw_i);   // (57)
    const Matrix3t& J_acc_angacc       = com_x;                                       // (56)
    const Matrix3t& J_angacc_torque    = I_inv;                                       // (53)
    const Matrix3t& J_angvelcom_angvel = -com_x;                                      // (39)
    const Matrix3t& J_angacc_angvel    = I_inv * (skew(I * _angvel) - angvel_x * I);  // (38)

    // Output Jacobians
    _J_acc_angvel =
        J_acc_angacc * J_angacc_angvel - angvel_x * J_angvelcom_angvel + skew(_angvel.cross(_com));  // (37)
    _J_acc_force   = Matrix<S, 3, 3>::Identity() / _mass;                                            // (59)
    _J_acc_torque  = J_acc_angacc * J_angacc_torque;                                                 // (?)
    _J_acc_com     = -angacc_x - angvel_x * angvel_x;                                                // (51)
    _J_acc_inertia = J_acc_angacc * J_angacc_inertia;                                                // (55)
    _J_acc_mass    = -_force / (_mass * _mass);                                                      // (50)
}

template <typename D1,
          typename D2,
          typename D3,
          typename D4,
          typename D5,
          typename D6,
          typename D7,
          typename D8,
          typename D9,
          typename D10,
          typename D11,
          typename D12,
          class T,
          class DT>
void tangent2delta(const MatrixBase<D1>& _tangent_acc_imu,
                   const MatrixBase<D2>& _tangent_angvel_imu,
                   const MatrixBase<D3>& _tangent_force,
                   const MatrixBase<D4>& _tangent_torque,
                   const MatrixBase<D5>& _com,
                   const MatrixBase<D6>& _inertia,
                   const T               _mass,
                   const DT              _dt,
                   MatrixBase<D7>&       _delta_p_imu,
                   MatrixBase<D8>&       _delta_v_imu,
                   MatrixBase<D9>&       _delta_p_dyn,
                   MatrixBase<D10>&      _delta_v_dyn,
                   MatrixBase<D11>&      _delta_L,
                   QuaternionBase<D12>&  _delta_q)
{
    DT dt2 = _dt * _dt;

    Matrix<typename D3::Scalar, 3, 1> acc_d;

    forces2acc(_tangent_force, _tangent_torque, _tangent_angvel_imu, _com, _inertia, _mass, acc_d);

    _delta_p_imu = 0.5 * _tangent_acc_imu * dt2;
    _delta_v_imu = _tangent_acc_imu * _dt;
    _delta_p_dyn = 0.5 * acc_d * dt2;
    _delta_v_dyn = acc_d * _dt;
    _delta_L     = _tangent_torque * _dt;
    _delta_q     = exp_q(_tangent_angvel_imu * _dt);
}

template <typename D1, typename D2, typename D3, typename D4, class T, class DT>
void tangent2delta(const MatrixBase<D1>& _tangent,
                   const MatrixBase<D2>& _com,
                   const MatrixBase<D3>& _inertia,
                   const T               _mass,
                   const DT              _dt,
                   MatrixBase<D4>&       _delta)
{
    // use shorter names
    Map<const Matrix<typename D1::Scalar, 3, 1>> acc_i(&_tangent(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> angvel(&_tangent(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> force(&_tangent(6));
    Map<const Matrix<typename D1::Scalar, 3, 1>> torque(&_tangent(9));
    Map<Matrix<typename D4::Scalar, 3, 1>>       _delta_p_imu(&_delta(0));
    Map<Matrix<typename D4::Scalar, 3, 1>>       _delta_v_imu(&_delta(3));
    Map<Matrix<typename D4::Scalar, 3, 1>>       _delta_p_dyn(&_delta(6));
    Map<Matrix<typename D4::Scalar, 3, 1>>       _delta_v_dyn(&_delta(9));
    Map<Matrix<typename D4::Scalar, 3, 1>>       _delta_L(&_delta(12));
    Map<Quaternion<typename D4::Scalar>>         _delta_q(&_delta(15));

    DT dt2 = _dt * _dt;

    Matrix<typename D1::Scalar, 3, 1> acc_d;

    forces2acc(force, torque, angvel, _com, _inertia, _mass, acc_d);

    _delta_p_imu = 0.5 * acc_i * dt2;
    _delta_v_imu = acc_i * _dt;
    _delta_p_dyn = 0.5 * acc_d * dt2;
    _delta_v_dyn = acc_d * _dt;
    _delta_L     = torque * _dt;
    _delta_q     = exp_q(angvel * _dt);
}

template <typename D1, typename D2, typename D3, class T>
void tangent2delta(const MatrixBase<D1>& _tangent, const MatrixBase<D2>& _model, const T _dt, MatrixBase<D3>& _delta)
{
    // use shorter names
    Map<const Matrix<typename D1::Scalar, 3, 1>> acc_i(&_tangent(0));
    Map<const Matrix<typename D1::Scalar, 3, 1>> angvel(&_tangent(3));
    Map<const Matrix<typename D1::Scalar, 3, 1>> force(&_tangent(6));
    Map<const Matrix<typename D1::Scalar, 3, 1>> torque(&_tangent(9));
    Map<const Matrix<typename D2::Scalar, 3, 1>> com(&_model(0));
    Map<const Matrix<typename D2::Scalar, 3, 1>> inertia(&_model(3));
    const typename D2::Scalar&                   mass(_model(6));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_p_imu(&_delta(0));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_v_imu(&_delta(3));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_p_dyn(&_delta(6));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_v_dyn(&_delta(9));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_L(&_delta(12));
    Map<Quaternion<typename D3::Scalar>>         _delta_q(&_delta(15));

    Matrix<typename D1::Scalar, 3, 1> acc_d;

    forces2acc(force, torque, angvel, com, inertia, mass, acc_d);

    T dt2 = _dt * _dt;

    // delta
    _delta_p_imu = 0.5 * acc_i * dt2;
    _delta_v_imu = acc_i * _dt;
    _delta_p_dyn = 0.5 * acc_d * dt2;
    _delta_v_dyn = acc_d * _dt;
    _delta_L     = torque * _dt;
    _delta_q     = exp_q(angvel * _dt);
}

template <typename D1, typename D2, typename D3, typename D4, typename D5, class T>
void tangent2delta(const MatrixBase<D1>& _tangent,
                   const MatrixBase<D2>& _model,
                   const T               _dt,
                   MatrixBase<D3>&       _delta,
                   MatrixBase<D4>&       _J_delta_tangent,
                   MatrixBase<D5>&       _J_delta_model)
{
    // blocks
    const unsigned int A  = 0;   // acc
    const unsigned int W  = 3;   // gyro
    const unsigned int F  = 6;   // force
    const unsigned int TQ = 9;   // torque
    const unsigned int C  = 0;   // CoM
    const unsigned int I  = 3;   // inertia
    const unsigned int M  = 6;   // mass
    const unsigned int PI = 0;   // position IMU
    const unsigned int VI = 3;   // velocity IMU
    const unsigned int PD = 6;   // position DYN
    const unsigned int VD = 9;   // velocity DYN
    const unsigned int L  = 12;  // ang momentum
    const unsigned int Q  = 15;  // orientation

    // use shorter names
    Map<const Matrix<typename D1::Scalar, 3, 1>> acc_i(&_tangent(A));
    Map<const Matrix<typename D1::Scalar, 3, 1>> angvel(&_tangent(W));
    Map<const Matrix<typename D1::Scalar, 3, 1>> force(&_tangent(F));
    Map<const Matrix<typename D1::Scalar, 3, 1>> torque(&_tangent(TQ));
    Map<const Matrix<typename D2::Scalar, 3, 1>> com(&_model(C));
    Map<const Matrix<typename D2::Scalar, 3, 1>> inertia(&_model(I));
    const typename D2::Scalar&                   mass(_model(M));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_p_imu(&_delta(PI));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_v_imu(&_delta(VI));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_p_dyn(&_delta(PD));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_v_dyn(&_delta(VD));
    Map<Matrix<typename D3::Scalar, 3, 1>>       _delta_L(&_delta(L));
    Map<Quaternion<typename D3::Scalar>>         _delta_q(&_delta(Q));

    Matrix<typename D1::Scalar, 3, 1> acc_d;

    Matrix<typename D1::Scalar, 3, 3> J_accd_force;
    Matrix<typename D1::Scalar, 3, 3> J_accd_torque;
    Matrix<typename D1::Scalar, 3, 3> J_accd_angvel;
    Matrix<typename D1::Scalar, 3, 3> J_accd_com;
    Matrix<typename D1::Scalar, 3, 3> J_accd_inertia;
    Matrix<typename D1::Scalar, 3, 1> J_accd_mass;

    forces2acc(force, torque, angvel, com, inertia, mass, 
               acc_d, 
               J_accd_force, J_accd_torque, J_accd_angvel,
               J_accd_com, J_accd_inertia, J_accd_mass);

    T dt2 = _dt * _dt;

    // delta
    _delta_p_imu = 0.5 * acc_i * dt2;
    _delta_v_imu = acc_i * _dt;
    _delta_p_dyn = 0.5 * acc_d * dt2;
    _delta_v_dyn = acc_d * _dt;
    _delta_q     = exp_q(angvel * _dt);
    _delta_L     = torque * _dt;

    // temporaries
    const Matrix<typename D1::Scalar, 3, 3>& I_dt    = Matrix<typename D1::Scalar, 3, 3>::Identity() * _dt;
    const Matrix<typename D1::Scalar, 3, 3>& I_dt2_2 = 0.5 * Matrix<typename D1::Scalar, 3, 3>::Identity() * dt2;

    // jacobians
    Matrix<typename D1::Scalar, 3, 3> J_dpi_acci  = I_dt2_2;
    Matrix<typename D1::Scalar, 3, 3> J_dvi_acci  = I_dt;
    const auto&                       J_dpd_accd  = I_dt2_2;
    const auto&                       J_dvd_accd  = I_dt;
    Matrix<typename D1::Scalar, 3, 3> J_dq_angvel = jac_SO3_right(angvel * _dt) * _dt;
    Matrix<typename D1::Scalar, 3, 3> J_dl_torque = I_dt;

    _J_delta_tangent.setZero();

    _J_delta_tangent.template block<3, 3>(PI, A) = J_dpi_acci;
    _J_delta_tangent.template block<3, 3>(VI, A) = J_dvi_acci;

    _J_delta_tangent.template block<3, 3>(PD, W)  = J_dpd_accd * J_accd_angvel;
    _J_delta_tangent.template block<3, 3>(PD, F)  = J_dpd_accd * J_accd_force;
    _J_delta_tangent.template block<3, 3>(PD, TQ) = J_dpd_accd * J_accd_torque;

    _J_delta_tangent.template block<3, 3>(VD, W)  = J_dvd_accd * J_accd_angvel;
    _J_delta_tangent.template block<3, 3>(VD, F)  = J_dvd_accd * J_accd_force;
    _J_delta_tangent.template block<3, 3>(VD, TQ) = J_dvd_accd * J_accd_torque;

    _J_delta_tangent.template block<3, 3>(Q, W)  = J_dq_angvel;
    _J_delta_tangent.template block<3, 3>(L, TQ) = J_dl_torque;

    _J_delta_model.setZero();

    _J_delta_model.template block<3, 3>(PD, C) = J_dpd_accd * J_accd_com;
    _J_delta_model.template block<3, 3>(PD, I) = J_dpd_accd * J_accd_inertia;
    _J_delta_model.template block<3, 1>(PD, M) = J_dpd_accd * J_accd_mass;

    _J_delta_model.template block<3, 3>(VD, C) = J_dvd_accd * J_accd_com;
    _J_delta_model.template block<3, 3>(VD, I) = J_dvd_accd * J_accd_inertia;
    _J_delta_model.template block<3, 1>(VD, M) = J_dvd_accd * J_accd_mass;
}

template <typename D1, typename D2, typename D3, typename D4, class T>
void data2delta(const MatrixBase<D1>&       _data,
                const MatrixBase<D2>&       _bias,
                const MatrixBase<D3>&       _model,
                const std::list<Propeller>& _propellers,
                const T                     _dt,
                MatrixBase<D4>&             _delta)
{
    Map<const Matrix<typename D2::Scalar, 3, 1>> com(&_model(0));

    Matrix<typename D4::Scalar, 12, 1> tangent;

    data2tangent(_data, _bias, com, _propellers, tangent);

    tangent2delta(tangent, _model, _dt, _delta);
}

template <typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, class T>
void data2delta(const MatrixBase<D1>&       _data,
                const MatrixBase<D2>&       _bias,
                const MatrixBase<D3>&       _model,
                const std::list<Propeller>& _propellers,
                const T                     _dt,
                MatrixBase<D4>&             _delta,
                MatrixBase<D5>&             _J_delta_data,
                MatrixBase<D6>&             _J_delta_bias,
                MatrixBase<D7>&             _J_delta_model)
{
    const unsigned int                           np = _propellers.size();
    Map<const Matrix<typename D2::Scalar, 3, 1>> com(&_model(0));

    Matrix<typename D4::Scalar, 12, 1> tangent;

    Matrix<typename D4::Scalar, 12, Dynamic> J_tangent_data(12, 6 + np);
    Matrix<typename D4::Scalar, 12, 12>      J_tangent_bias;
    Matrix<typename D4::Scalar, 12, 7>       J_tangent_model;
    Matrix<typename D4::Scalar, 18, 12>      J_delta_tangent;

    J_tangent_model.setZero(12, 7);

    Matrix<typename D4::Scalar, 12, 3> J_tangent_com;
    data2tangent(_data, _bias, com, _propellers, tangent, J_tangent_data, J_tangent_bias, J_tangent_com);
    J_tangent_model.template block<12, 3>(0, 0) = J_tangent_com;

    tangent2delta(tangent, _model, _dt, _delta, J_delta_tangent, _J_delta_model);

    _J_delta_data  = J_delta_tangent * J_tangent_data;
    _J_delta_bias  = J_delta_tangent * J_tangent_bias;
    _J_delta_model = _J_delta_model + J_delta_tangent * J_tangent_model;
}


}  // namespace fti
}  // namespace bodydynamics
}  // namespace wolf

#endif /* MATH_FORCE_TORQUE_INERTIAL_DELTA_TOOLS_H_ */
