//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef SENSOR_FORCE_TORQUE_INERTIAL_H
#define SENSOR_FORCE_TORQUE_INERTIAL_H

// wolf includes
#include "core/sensor/sensor_base.h"
#include "core/utils/params_server.h"
#include <iostream>

namespace wolf
{

WOLF_STRUCT_PTR_TYPEDEFS(ParamsSensorForceTorqueInertial);

struct ParamsSensorForceTorqueInertial : public ParamsSensorBase
{
    // noise std dev
    double   acc_noise_std;
    double   gyro_noise_std;
    double   accb_initial_std;
    double   gyrob_initial_std;
    double   acc_drift_std;
    double   gyro_drift_std;
    double   force_noise_std;
    double   torque_noise_std;
    Vector3d com;
    Vector3d inertia;
    double   mass;

    ParamsSensorForceTorqueInertial() = default;
    ParamsSensorForceTorqueInertial(std::string _unique_name, const ParamsServer& _server)
        : ParamsSensorBase(_unique_name, _server)
    {
        acc_noise_std     = _server.getParam<double>(prefix + _unique_name + "/acc_noise_std");
        gyro_noise_std    = _server.getParam<double>(prefix + _unique_name + "/gyro_noise_std");
        accb_initial_std  = _server.getParam<double>(prefix + _unique_name + "/accb_initial_std");
        gyrob_initial_std = _server.getParam<double>(prefix + _unique_name + "/gyrob_initial_std");
        acc_drift_std     = _server.getParam<double>(prefix + _unique_name + "/acc_drift_std");
        gyro_drift_std    = _server.getParam<double>(prefix + _unique_name + "/gyro_drift_std");
        force_noise_std   = _server.getParam<double>(prefix + _unique_name + "/force_noise_std");
        torque_noise_std  = _server.getParam<double>(prefix + _unique_name + "/torque_noise_std");
        com               = _server.getParam<Vector3d>(prefix + _unique_name + "/com");
        inertia           = _server.getParam<Vector3d>(prefix + _unique_name + "/inertia");
        mass              = _server.getParam<double>(prefix + _unique_name + "/mass");
    }
    ~ParamsSensorForceTorqueInertial() override = default;
    std::string print() const override
    {
        return ParamsSensorBase::print() + "\n"                                      //
               + "acc_noise_std:    " + std::to_string(acc_noise_std) + "\n"         //
               + "gyro_noise_std:   " + std::to_string(gyro_noise_std) + "\n"        //
               + "accb_initial_std:    " + std::to_string(accb_initial_std) + "\n"   //
               + "gyrob_initial_std:   " + std::to_string(gyrob_initial_std) + "\n"  //
               + "acc_drift_std:    " + std::to_string(acc_drift_std) + "\n"         //
               + "gyro_drift_std:   " + std::to_string(gyro_drift_std) + "\n"        //
               + "force_noise_std:  " + std::to_string(force_noise_std) + "\n"       //
               + "torque_noise_std: " + std::to_string(torque_noise_std) + "\n"      //
               + "com: print not implemented." + "\n"                                //
               + "inertia: print not implemented. \n"                                //
               + "mass:             " + std::to_string(mass) + "\n";                 //
    }
};

WOLF_PTR_TYPEDEFS(SensorForceTorqueInertial);

class SensorForceTorqueInertial : public SensorBase
{
  protected:
    ParamsSensorForceTorqueInertialPtr params_fti_;

  public:
    SensorForceTorqueInertial(const VectorComposite& _states, ParamsSensorForceTorqueInertialPtr _params);
    SensorForceTorqueInertial(const VectorXd& _extrinsics, ParamsSensorForceTorqueInertialPtr _params);
    ~SensorForceTorqueInertial() override = default;

    WOLF_SENSOR_CREATE(SensorForceTorqueInertial, ParamsSensorForceTorqueInertial, 7);

    // getters return by copy
    Vector6d                           getImuBias() const;   // Imu bias [acc, gyro]
    Vector3d                           getAccBias() const;   // Acc bias
    Vector3d                           getGyroBias() const;  // Gyro bias
    double                             getMass() const;      // Total mass
    Vector3d                           getInertia() const;   // Inertia vector (diagonal of inertia matrix)
    Vector3d                           getCom() const;       // centre of mass
    Vector7d                           getModel() const;     // dynamic model [com, inertia, mass]
    ParamsSensorForceTorqueInertialPtr getParamsSensorForceTorqueInertial()
    {
        return params_fti_;
    }
};

inline double SensorForceTorqueInertial::getMass() const
{
    return getStateBlock('m')->getState()(0);
}

inline Vector6d SensorForceTorqueInertial::getImuBias() const
{
    return getStateBlock('I')->getState();
}

inline Vector3d SensorForceTorqueInertial::getAccBias() const
{
    return getStateBlock('I')->getState().head<3>();
}

inline Vector3d SensorForceTorqueInertial::getGyroBias() const
{
    return getStateBlock('I')->getState().tail<3>();
}

inline Vector3d SensorForceTorqueInertial::getInertia() const
{
    return getStateBlock('i')->getState();
}

inline Vector3d SensorForceTorqueInertial::getCom() const
{
    return getStateBlock('C')->getState();
}

inline Vector7d SensorForceTorqueInertial::getModel() const
{
    Vector7d model;
    model << getCom(), getInertia(), getMass();
    return model;
}

}  // namespace wolf

#endif  // SENSOR_FORCE_TORQUE_H
