//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef SENSOR_POINT_FEET_NOMOVE_H
#define SENSOR_POINT_FEET_NOMOVE_H

//wolf includes
#include "core/sensor/sensor_base.h"
#include "core/utils/params_server.h"
#include <iostream>

namespace wolf {

WOLF_STRUCT_PTR_TYPEDEFS(ParamsSensorPointFeetNomove);


struct ParamsSensorPointFeetNomove : public ParamsSensorBase
{
        // standard deviation on the assumption that the feet are not moving when in contact
        // the noise represents a random walk on the foot position, hence the unit 
        double std_foot_nomove_; // m/(s^2 sqrt(dt))
        // certainty on current ground altitude
        double std_altitude_ = 1000; // m, deactivated by default
        double foot_radius_ = 0; // m

        ParamsSensorPointFeetNomove() = default;
        ParamsSensorPointFeetNomove(std::string _unique_name, const ParamsServer& _server):
            ParamsSensorBase(_unique_name, _server)
        {
            std_foot_nomove_ = _server.getParam<double>(_unique_name + "/std_foot_nomove");
            std_altitude_ = _server.getParam<double>(_unique_name + "/std_altitude");
            foot_radius_ = _server.getParam<double>(_unique_name + "/foot_radius");
        }
        ~ParamsSensorPointFeetNomove() override = default;
        std::string print() const override
        {
            return "\n" + ParamsSensorBase::print()                          + "\n"
                    + "std_altitude:    " + std::to_string(std_altitude_)    + "\n"
                    + "std_foot_nomove: " + std::to_string(std_foot_nomove_) + "\n";
                    + "foot_radius_:    " + std::to_string(foot_radius_) + "\n";
                    
        }
};

WOLF_PTR_TYPEDEFS(SensorPointFeetNomove);

class SensorPointFeetNomove : public SensorBase
{

    protected:
        Matrix3d cov_foot_nomove_;  // random walk covariance in (m/s^2/sqrt(Hz))^2
        Matrix1d cov_altitude_;  // m^2
        double foot_radius_; 

    public:

        SensorPointFeetNomove(const Eigen::VectorXd& _extrinsics, const ParamsSensorPointFeetNomovePtr& _intr_pfnm);
        ~SensorPointFeetNomove() override;

        WOLF_SENSOR_CREATE(SensorPointFeetNomove, ParamsSensorPointFeetNomove, 0);

        Matrix3d getCovFootNomove() const;
        Matrix1d getCovAltitude() const;
        double getFootRadius() const;
};

inline Matrix3d SensorPointFeetNomove::getCovFootNomove() const
{
    return cov_foot_nomove_;
}

inline Matrix1d SensorPointFeetNomove::getCovAltitude() const
{
    return cov_altitude_;
}

inline double SensorPointFeetNomove::getFootRadius() const
{
    return foot_radius_;
}


} // namespace wolf

#endif // SENSOR_POINT_FEET_NOMOVE_H
