//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * processor_force_torque_inertial.cpp
 *
 *  Created on: Aug 19, 2021
 *      Author: jsola
 */

// bodydynamics
#include "bodydynamics/processor/processor_force_torque_inertial.h"
#include "bodydynamics/factor/factor_force_torque_inertial.h"
#include "bodydynamics/math/force_torque_inertial_delta_tools.h"

// wolf
#include <core/state_block/state_composite.h>
#include <core/capture/capture_motion.h>
#include <core/factor/factor_block_difference.h>

namespace wolf
{

ProcessorForceTorqueInertial::ProcessorForceTorqueInertial(
    ParamsProcessorForceTorqueInertialPtr _params_force_torque_inertial)
    : ProcessorMotion("ProcessorForceTorqueInertial",
                      "POVL",
                      3,   // dim
                      13,  // state size [p, q, v, L]
                      19,  // delta size [pi, vi, pd, vd, L, q]
                      18,  // delta cov size
                      12,  // data size [a, w, f, t]
                      6,   // calib size [ab, wb]
                      _params_force_torque_inertial),
      params_force_torque_inertial_(_params_force_torque_inertial)
{
    // TODO Auto-generated constructor stub
}

ProcessorForceTorqueInertial::~ProcessorForceTorqueInertial()
{
    // TODO Auto-generated destructor stub
}

CaptureMotionPtr ProcessorForceTorqueInertial::emplaceCapture(const FrameBasePtr&   _frame_own,
                                                              const SensorBasePtr&  _sensor,
                                                              const TimeStamp&      _ts,
                                                              const VectorXd&       _data,
                                                              const MatrixXd&       _data_cov,
                                                              const VectorXd&       _calib,
                                                              const VectorXd&       _calib_preint,
                                                              const CaptureBasePtr& _capture_origin_ptr)
{
    CaptureMotionPtr capture = CaptureBase::emplace<CaptureMotion>(
        _frame_own, "CaptureForceTorqueInertial", _ts, _sensor, _data, _data_cov, _capture_origin_ptr);
    setCalibration(capture, _calib);
    capture->setCalibrationPreint(_calib_preint);
    return capture;
}

void ProcessorForceTorqueInertial::emplaceFeaturesAndFactors(CaptureBasePtr   _capture_origin,
                                                                     CaptureMotionPtr _capture_own)
{
    // 1. Feature and factor FTI -- force-torque-inertial
    //----------------------------------------------------
    auto motion  = _capture_own->getBuffer().back();
    auto ftr_fti = FeatureBase::emplace<FeatureMotion>(_capture_own,
                                                       "FeatureMotion",
                                                       motion.delta_integr_,
                                                       motion.delta_integr_cov_ + unmeasured_perturbation_cov_,
                                                       _capture_own->getCalibrationPreint(),
                                                       motion.jacobian_calib_);

    FactorBase::emplace<FactorForceTorqueInertial>(
        ftr_fti, ftr_fti, _capture_origin, shared_from_this(), params_->apply_loss_function);


    // 2. Feature and factor bias -- IMU bias drift for acc and gyro
    //---------------------------------------------------------------
    // - This factor only if IMU biases are Dynamic (check sensor->isStateBlockDynamic('I'))
    // - feature has zero measurement size 6, with the cov of the bias drift size 6x6
    // - factor relates bias(last capture) and bias(origin capture)
    if (getSensor()->isStateBlockDynamic('I'))
    {
        const auto& sb_imubias_own    = _capture_own->getStateBlock('I');
        const auto& sb_imubias_origin = _capture_origin->getStateBlock('I');
        if (sb_imubias_own != sb_imubias_origin)  // make sure it's two different state blocks! -- just in case
        {
            auto dt = _capture_own->getTimeStamp() - _capture_origin->getTimeStamp();
            auto ftr_bias =
                FeatureBase::emplace<FeatureBase>(_capture_own,
                                                  "FeatureBase",
                                                  Vector6d::Zero(),      // mean IMU drift is zero
                                                  imu_drift_cov_ * dt);  // IMU drift cov specified in continuous time
            FactorBase::emplace<FactorBlockDifference>(ftr_bias,
                                                       ftr_bias,
                                                       sb_imubias_own,      // IMU bias block at t=own
                                                       sb_imubias_origin,   // IMU bias block at t=origin
                                                       nullptr,             // frame other
                                                       _capture_origin,     // origin capture
                                                       nullptr,             // feature other
                                                       nullptr,             // landmark other
                                                       0,                   // take all of first state block
                                                       -1,                  // take all of first state block
                                                       0,                   // take all of first second block
                                                       -1,                  // take all of first second block
                                                       shared_from_this(),  // this processor
                                                       false);              // loss function
        }
    }
}


void ProcessorForceTorqueInertial::setCalibration(const CaptureBasePtr _capture, const VectorXd& _calibration)
{
    _capture->getStateBlock('I')->setState(_calibration);  // Set IMU bias
}

void ProcessorForceTorqueInertial::configure(SensorBasePtr _sensor)
{
    sensor_fti_ = std::static_pointer_cast<SensorForceTorqueInertial>(_sensor);


    auto acc_drift_std  = sensor_fti_->getParamsSensorForceTorqueInertial()->acc_drift_std;
    auto gyro_drift_std = sensor_fti_->getParamsSensorForceTorqueInertial()->gyro_drift_std;

    ArrayXd imu_drift_sigmas(6);
    imu_drift_sigmas << acc_drift_std, acc_drift_std, acc_drift_std, gyro_drift_std, gyro_drift_std, gyro_drift_std;
    imu_drift_cov_  = imu_drift_sigmas.square().matrix().asDiagonal();
}

void ProcessorForceTorqueInertial::computeCurrentDelta(const Eigen::VectorXd& _data,
                                                       const Eigen::MatrixXd& _data_cov,
                                                       const Eigen::VectorXd& _calib,
                                                       const double           _dt,
                                                       Eigen::VectorXd&       _delta,
                                                       Eigen::MatrixXd&       _delta_cov,
                                                       Eigen::MatrixXd&       _jacobian_calib) const
{
    // compute tangent by removing IMU bias
    Matrix<double, 12, 1> tangent = _data;
    tangent.head<6>() -= _calib;            // J_tangent_data  = Identity(12,12)
    Matrix<double, 12, 6> J_tangent_calib;  // J_tangent_calib = [-Identity(6,6) ; Zero(6,6)]
    J_tangent_calib.topRows<6>()    = -Matrix6d::Identity();
    J_tangent_calib.bottomRows<6>() = Matrix6d::Zero();

    // go from tangent to delta. We need to dynamic model for this
    const Matrix<double, 7, 1>& model = sensor_fti_->getModel();
    Matrix<double, 18, 12>      J_delta_tangent;
    Matrix<double, 18, 7>       J_delta_model;
    fti::tangent2delta(tangent, model, _dt, _delta, J_delta_tangent, J_delta_model);

    // Compute cov and compose jacobians
    const auto& J_delta_data = J_delta_tangent;  //  * J_tangent_data, which is the Identity;
    _delta_cov               = J_delta_data * _data_cov * J_delta_data.transpose();
    _jacobian_calib          = J_delta_tangent * J_tangent_calib;
}

void ProcessorForceTorqueInertial::deltaPlusDelta(const Eigen::VectorXd& _delta1,
                                                  const Eigen::VectorXd& _delta2,
                                                  const double           _dt2,
                                                  Eigen::VectorXd&       _delta1_plus_delta2) const
{
    bodydynamics::fti::compose(_delta1, _delta2, _dt2, _delta1_plus_delta2);
}

void ProcessorForceTorqueInertial::deltaPlusDelta(const Eigen::VectorXd& _delta1,
                                                  const Eigen::VectorXd& _delta2,
                                                  const double           _dt2,
                                                  Eigen::VectorXd&       _delta1_plus_delta2,
                                                  Eigen::MatrixXd&       _jacobian1,
                                                  Eigen::MatrixXd&       _jacobian2) const
{
    bodydynamics::fti::compose(_delta1, _delta2, _dt2, _delta1_plus_delta2, _jacobian1, _jacobian2);
}

void ProcessorForceTorqueInertial::statePlusDelta(const VectorComposite& _x,
                                                  const Eigen::VectorXd& _delta,
                                                  const double           _dt,
                                                  VectorComposite&       _x_plus_delta) const
{
    auto     x            = _x.vector("PVLO");
    VectorXd x_plus_delta = bodydynamics::fti::composeOverState(x, _delta, _dt);
    _x_plus_delta         = VectorComposite(x_plus_delta, "PVLO", {3, 3, 3, 4});
}

Eigen::VectorXd ProcessorForceTorqueInertial::correctDelta(const Eigen::VectorXd& _delta_preint,
                                                           const Eigen::VectorXd& _delta_step) const
{
    return fti::plus(_delta_preint, _delta_step);
}

VectorXd ProcessorForceTorqueInertial::getCalibration(const CaptureBaseConstPtr _capture) const
{
    if (_capture)
        return _capture->getStateBlock('I')->getState();  // IMU bias
    else
        return getSensor()->getStateBlock('I')->getState();  // IMU bias
}
} /* namespace wolf */

#include <core/processor/factory_processor.h>
namespace wolf
{
WOLF_REGISTER_PROCESSOR(ProcessorForceTorqueInertial);
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorForceTorqueInertial);
}  // namespace wolf