//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * processor_force_torque_inertial_dynamics.cpp
 *
 *  Created on: Aug 1, 2022
 *      Author: asanjuan
 */

// bodydynamics
#include "bodydynamics/processor/processor_force_torque_inertial_dynamics.h"
#include "bodydynamics/factor/factor_angular_momentum.h"
#include "bodydynamics/factor/factor_force_torque_inertial_dynamics.h"
#include "bodydynamics/math/force_torque_inertial_delta_tools.h"
#include "bodydynamics/capture/capture_force_torque_inertial.h"

// wolf
#include <core/state_block/state_composite.h>
#include <core/capture/capture_motion.h>
#include <core/factor/factor_block_difference.h>

namespace wolf
{

ProcessorForceTorqueInertialDynamics::ProcessorForceTorqueInertialDynamics(
    ParamsProcessorForceTorqueInertialDynamicsPtr _params_force_torque_inertial_dynamics)
    : ProcessorMotion("ProcessorForceTorqueInertialDynamics",
                      "POVL",
                      3,   // dim
                      13,  // state size [p, q, v, L]
                      19,  // delta size [pi, vi, pd, vd, L, q]
                      18,  // delta cov size
                      12,  // data size [a, w, f, t]
                      13,  // calib size [ab, wb, c, i, m]
                      _params_force_torque_inertial_dynamics),
      params_force_torque_inertial_dynamics_(_params_force_torque_inertial_dynamics)
{
    //
}

ProcessorForceTorqueInertialDynamics::~ProcessorForceTorqueInertialDynamics()
{
    //
}

CaptureMotionPtr ProcessorForceTorqueInertialDynamics::emplaceCapture(const FrameBasePtr&   _frame_own,
                                                                      const SensorBasePtr&  _sensor,
                                                                      const TimeStamp&      _ts,
                                                                      const VectorXd&       _data,
                                                                      const MatrixXd&       _data_cov,
                                                                      const VectorXd&       _calib,
                                                                      const VectorXd&       _calib_preint,
                                                                      const CaptureBasePtr& _capture_origin_ptr)
{
    CaptureMotionPtr capture = CaptureBase::emplace<CaptureForceTorqueInertial>(
        _frame_own, _ts, _sensor, _data, _data_cov, _capture_origin_ptr);
    setCalibration(capture, _calib);
    capture->setCalibrationPreint(_calib_preint);
    return capture;
}

void ProcessorForceTorqueInertialDynamics::emplaceFeaturesAndFactors(CaptureBasePtr   _capture_origin,
                                                                     CaptureMotionPtr _capture_own)
{
    // 1. Feature and factor FTI -- force-torque-inertial
    //----------------------------------------------------
    auto motion  = _capture_own->getBuffer().back();
    auto ftr_fti = FeatureBase::emplace<FeatureMotion>(_capture_own,
                                                       "FeatureMotion",
                                                       motion.delta_integr_,
                                                       motion.delta_integr_cov_ + unmeasured_perturbation_cov_,
                                                       _capture_own->getCalibrationPreint(),
                                                       motion.jacobian_calib_);

    FactorBase::emplace<FactorForceTorqueInertialDynamics>(
        ftr_fti, ftr_fti, _capture_origin, shared_from_this(), params_->apply_loss_function);

    // 2. Feature and factor L -- angular momentum
    //---------------------------------------------
    // - feature has the current gyro measurement
    // - factor relates w_measured, IMU_bias(last capture), L(last capture) and i(sensor)

    auto measurement_gyro = motion.data_.segment<3>(3);
    auto gyro_cov         = motion.data_cov_.block<3, 3>(3, 3);

    auto ftr_base = FeatureBase::emplace<FeatureBase>(_capture_own, "FeatureBase", measurement_gyro, gyro_cov);

    FactorBase::emplace<FactorAngularMomentum>(ftr_base, ftr_base, shared_from_this(), params_->apply_loss_function);

    // 3. Feature and factor bias -- IMU bias drift for acc and gyro
    //---------------------------------------------------------------
    // - This factor only if IMU biases are Dynamic (check sensor->isStateBlockDynamic('I'))
    // - feature has zero measurement size 6, with the cov of the bias drift size 6x6
    // - factor relates bias(last capture) and bias(origin capture)
    if (getSensor()->isStateBlockDynamic('I'))
    {
        const auto& sb_imubias_own    = _capture_own->getStateBlock('I');
        const auto& sb_imubias_origin = _capture_origin->getStateBlock('I');
        if (sb_imubias_own != sb_imubias_origin)  // make sure it's two different state blocks! -- just in case
        {
            auto dt = _capture_own->getTimeStamp() - _capture_origin->getTimeStamp();
            auto ftr_bias =
                FeatureBase::emplace<FeatureBase>(_capture_own,
                                                  "FeatureBase",
                                                  Vector6d::Zero(),      // mean IMU drift is zero
                                                  imu_drift_cov_ * dt);  // IMU drift cov specified in continuous time
            FactorBase::emplace<FactorBlockDifference>(ftr_bias,
                                                       ftr_bias,
                                                       sb_imubias_own,      // IMU bias block at t=own
                                                       sb_imubias_origin,   // IMU bias block at t=origin
                                                       nullptr,             // frame other
                                                       _capture_origin,     // origin capture
                                                       nullptr,             // feature other
                                                       nullptr,             // landmark other
                                                       0,                   // take all of first state block
                                                       -1,                  // take all of first state block
                                                       0,                   // take all of first second block
                                                       -1,                  // take all of first second block
                                                       shared_from_this(),  // this processor
                                                       params_->apply_loss_function); // loss function
        }
    }
}

void ProcessorForceTorqueInertialDynamics::setCalibration(const CaptureBasePtr _capture, const VectorXd& _calibration)
{
    _capture->getStateBlock('I')->setState(_calibration.segment<6>(0));   // Set IMU bias
    _capture->getStateBlock('C')->setState(_calibration.segment<3>(6));   // Set C
    _capture->getStateBlock('i')->setState(_calibration.segment<3>(9));   // Set i
    _capture->getStateBlock('m')->setState(_calibration.segment<1>(12));  // Set m
}

void ProcessorForceTorqueInertialDynamics::configure(SensorBasePtr _sensor)
{
    sensor_fti_ = std::static_pointer_cast<SensorForceTorqueInertial>(_sensor);

    auto acc_drift_std  = sensor_fti_->getParamsSensorForceTorqueInertial()->acc_drift_std;
    auto gyro_drift_std = sensor_fti_->getParamsSensorForceTorqueInertial()->gyro_drift_std;

    ArrayXd imu_drift_sigmas(6);
    imu_drift_sigmas << acc_drift_std, acc_drift_std, acc_drift_std, gyro_drift_std, gyro_drift_std, gyro_drift_std;
    imu_drift_cov_  = imu_drift_sigmas.square().matrix().asDiagonal();
}

void ProcessorForceTorqueInertialDynamics::data2tangent(const Eigen::VectorXd& _data,
                                                        const Eigen::VectorXd& _bias,
                                                        const Eigen::VectorXd& _model,
                                                        Eigen::VectorXd&       _tangent,
                                                        Eigen::MatrixXd&       _J_tangent_data,
                                                        Eigen::MatrixXd&       _J_tangent_bias,
                                                        Eigen::MatrixXd&       _J_tangent_model) const

{
    const Vector6d& awd      = _data.segment<6>(0);  // this is (a,w) of data in a 6-vector
    const Vector3d& fd       = _data.segment<3>(6);
    const Vector3d& td       = _data.segment<3>(9);
    const Vector3d& c        = _model.segment<3>(0);
    const Matrix3d& fd_cross = skew(fd);
    const Matrix3d& c_cross  = skew(c);

    _tangent.segment<6>(0) = awd - _bias;       // awt = awd - awb
    _tangent.segment<3>(6) = fd;                // ft = fd
    _tangent.segment<3>(9) = td - c.cross(fd);  // tt = td - c x fd

    // J_tangent_data
    _J_tangent_data.setIdentity(12, 12);
    _J_tangent_data.block<3, 3>(9, 6) = -c_cross;  // J_tt_fd = -c_cross

    // J_tangent_bias = [-Identity(6,6) ; Zero(6,6)]
    _J_tangent_bias.topRows<6>()    = -Matrix6d::Identity();
    _J_tangent_bias.bottomRows<6>() = Matrix6d::Zero();

    // J_tangent_model
    _J_tangent_model.setZero(12, 7);
    _J_tangent_model.block<3, 3>(9, 0) = fd_cross;  // J_tt_c = fd_cross
}

void ProcessorForceTorqueInertialDynamics::computeCurrentDelta(const Eigen::VectorXd& _data,
                                                               const Eigen::MatrixXd& _data_cov,
                                                               const Eigen::VectorXd& _calib,
                                                               const double           _dt,
                                                               Eigen::VectorXd&       _delta,
                                                               Eigen::MatrixXd&       _delta_cov,
                                                               Eigen::MatrixXd&       _jacobian_calib) const
{
    // compute tangent by removing IMU bias

    /**
     * Notes on the computation of `tangent` and `delta` from `data` and `calib`, and its Jacobians:
     * =============================================================================================
     *
     * Overview: from sensor data and system calibration parameters, we want to compute a delta.
     * We proceed in two steps:
     *   - tangent  = f ( data, bias, model )
     *   - delta    = g (tangent, model, dt )
     *
     * Notice that the `model` part appears in both stages of the computation, f() and g().
     * This means that the chain rule for `J_delta_model` will have a double path.
     *
     * 1) On a first stage, we want to obtain `tangent`
     * ------------------------------------------------
     * We have the following partitions of our variables:
     *
     *  tangent  = [at,wt,ft,tt]    -- tangent to the delta manifold. (a,w) at BASE, (f,t) at COM
     *  data     = [ad,wd,fd,td]    -- sensor data, all measured at BASE
     *  bias     = [ab,wb]          -- IMU bias. This we called `I` in previous versions
     *  model    = [c,i,m]          -- system dynamic model, where c = r_base_com (CoM wrt BASE)
     *  calib    = [bias,model]     -- system calibration parameters
     *
     * we have the function:
     *
     *  tangent = f (data, bias, model)
     *
     * we need to obtain:
     *  - tangent
     *  - J_tangent_data
     *  - J_tangent_bias
     *  - J_tangent_model
     *
     * We express each one of the blocks of `tangent` wrt the blocks of `data`, `bias` and `model`:
     *
     *  at = ad - ab        --> J_at_ad = I33 ; J_at_ab = -I33
     *  wt = wd - wb        --> J_wt_wd = I33 ; J_wt_wb = -I33
     *  ft = fd             --> J_ft_fd = I33
     *  tt = td - c x fd    --> J_tt_td = I33 ; J_tt_c = skew(fd) = fd_x ; J_tt_fd = -skew(c) = -c_x
     *
     * note: to obtain `tt` (at COM) we need to account for the torque measurement `td` (at BASE)
     * and add the torque produced by the force `fd` applied at the lever arm of BASE wrt COM, which is `-c`,
     * obtaining `tt = td - c x fd`
     *
     * So we can assemble the tangent vector as
     *
     *  tangent = [at,wt,ft,tt]
     *
     * and all the Jacobian blocks as:
     *
     *                      ad      wd      fd      td
     *  J_tangent_data = [  I33     033     033     033     at
     *                      033     I33     033     033     wt
     *                      033     033     I33     033     ft
     *                      033     033     -c_x    I33 ]   tt
     *
     *                      ab     wb
     *  J_tangent_bias = [  -I33    033     at
     *                      033     -I33    wt
     *                      033     033     ft
     *                      033     033 ]   tt
     *
     *                       c      i       m
     *  J_tangent_model = [  033    033     031     at
     *                       033    033     031     wt
     *                       033    033     031     ft
     *                       fd_x   033     031 ]   tt
     *
     * and from here on, we just need to fill in the matrices.
     *
     * Notations:
     *  I33 : 3x3 Identity matrix
     *  033 : 3x3 zero matrix
     *  031 : 3x1 zero vector
     *
     *
     * 2) On a second stage, we need to obtain the `delta` from the previous results.
     * ------------------------------------------------------------------------------
     *
     * We have:
     *
     *  delta = g (tangent, model, dt)
     *
     * we need to obtain:
     *
     *  delta
     *  J_delta_tangent
     *  J_delta_model
     *
     * these are obtained directly by the function g() = tangent2delta(). This function does:
     *
     *  angacc      = I.inv * ( tt -  wt x ( I * wt ) )    
     *      --> J_aa_i, J_aa_tt, J_aa_wt
     *  acc_dyn     = ft / _mass - angacc x com - wt x ( wt x com )  
     *      --> J_ad_ft, J_ad_m, J_ad_aa, J_ad_c, J_ad_wt
     *      ==> J_ad_wt, J_ad_ft, J_ad_tt, J_ad_c, J_ad_i, J_ad_m
     *
     *  delta_p_imu = 0.5 * at * dt^2       --> J_dpi_at
     *  delta_v_imu = at * _dt              --> J_dvi_at
     *  delta_p_dyn = 0.5 * acc_dyn * dt^2  --> J_dpd_ad
     *  delta_v_dyn = acc_dyn * _dt         --> J_dvd_ad
     *  delta_L     = tt * _dt              --> J_dL_tt
     *  delta_q     = exp_q(wt * _dt)       --> J_dq_wt
     * 
     * Assembling Jacobians gives:
     * 
     *                      at          wt          ft          tt
     *  J_delta_tangent = [ J_dpi_at    0           0           0
     *                      J_dvi_at    0           0           0
     *                      0           J_dpd_wt    J_dpd_ft    J_dpd_tt
     *                      0           J_dvd_wt    J_dvd_ft    J_dvd_tt
     *                      0           0           0           J_dL_tt
     *                      0           J_dq_wt     0           0       ]
     * with
     *  J_dpd_wt = J_dpd_ad * ( J_ad_wt + J_ad_aa * J_aa_wt )
     *  J_dpd_ft = J_dpd_ad * J_ad_ft
     *  J_dpd_tt = J_dpd_ad * J_ad_aa * J_aa_tt
     *  J_dvd_wt = J_dvd_ad * ( J_ad*wt + J_ad_aa * J_aa_wt )
     *  J_dvd_ft = J_dvd_ad * J_ad_ft
     *  J_dvd_tt = J_dvd_ad * J_ad_aa * J_aa_tt
     *
     *                      c       i       m
     *  J_delta_model = [   0       0       0
     *                      0       0       0
     *                      J_dpd_c J_dpd_i J_dpd_m
     *                      J_dvd_c J_dvd_i J_dvd_m
     *                      0       0       0
     *                      0       0       0   ]
     * with
     *  J_dpd_c = J_dpd_ad * J_ad_c
     *  J_dpd_i = J_dpd_ad * J_ad_aa * J_aa_i
     *  J_dpd_m = J_dpd_ad * J_ad_m
     *  J_dvd_c = J_dvd_ad * J_ad_c
     *  J_dvd_i = J_dvd_ad * J_ad_aa * J_aa_i
     *  J_dvd_m = J_dvd_ad * J_ad_m
     *
     * 3) On a third stage, we need to apply the chain rule for the functions f() and g(),
     * that is delta = g( f( data, bias, model), model)  -- yes, it's a mess!!
     *
     *  J_delta_data  = J_delta_tangent * J_tangent_data
     *  J_delta_model = J_delta_model   + J_delta_tangent * J_tangent_model <-- remember all Jacobians are PARTIAL
     * derivatives.
     *  J_delta_bias  = J_delta_tangent * J_tangent_bias
     *
     * Finally we can assemble the Jacobian `J_delta_calib` as the concatenation of bias and model:
     *
     *  J_delta_calib = [ J_delta_bias | J_delta_model ]
     *
     *
     * 4) On a fourth stage, we compute the covariance of the delta:
     *
     *  Q_delta = J_delta_data * Q_data * J_delta_data.transpose()
     *
     *
     * 5) The function returns the following quantities, which relate to the variables used above:
     *
     *  - _delta            = delta
     *  - _delta_cov        = Q_delta
     *  - _jacobian_calib   = J_delta_calib
     *
     * ---------------------------------------------------------------
     *
     * Notes: Jacobians of cross product
     *
     *  a, b in R^3
     *
     *  a_x, b_x, skew-symmetric matrices, in R^3x3
     *
     *  a_x = [0 -az ay ; az 0 -ax ; -ay ax 0]
     *
     *  a x b = a_x b
     *
     *  a x b = - b x a = - b_x a
     *
     *  J_(axb)_a = -b_x
     *
     *  J_(axb)_b = a_x
     */

    /*
     * 1. tangent = f(data, bias, model) --> J_tangent_data, J_tangent_bias, J_tangent_model
     *
     * data     = [ad,wd,fd,td]
     * bias     = [ab,wb]
     * model    = [c,i,m]
     * tangent  = [at,wt,ft,tt]
     */

    Vector6d bias  = _calib.segment<6>(0);
    Vector7d model = _calib.segment<7>(6);
    VectorXd tangent(12);
    MatrixXd J_tangent_data(12, 12);
    MatrixXd J_tangent_bias(12, 6);
    MatrixXd J_tangent_model(12, 7);

    data2tangent(_data, bias, model, tangent, J_tangent_data, J_tangent_bias, J_tangent_model);

    // 2. go from tangent to delta. We need again the dynamic model for this
    Matrix<double, 18, 12> J_delta_tangent;
    Matrix<double, 18, 7>  J_delta_model;
    fti::tangent2delta(
        tangent, model, _dt, _delta, J_delta_tangent, J_delta_model);  

    // 3. Compose jacobians
    Matrix<double, 18, 6> J_delta_bias = J_delta_tangent * J_tangent_bias;
    J_delta_model += J_delta_tangent * J_tangent_model;
    _jacobian_calib << J_delta_bias, J_delta_model;  // J_delta_calib = [J_delta_bias ; J_delta_model]
    const auto& J_delta_data = J_delta_tangent * J_tangent_data;

    // 4. compute covariance
    _delta_cov = J_delta_data * _data_cov * J_delta_data.transpose();

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////
}

void ProcessorForceTorqueInertialDynamics::deltaPlusDelta(const Eigen::VectorXd& _delta1,
                                                          const Eigen::VectorXd& _delta2,
                                                          const double           _dt2,
                                                          Eigen::VectorXd&       _delta1_plus_delta2) const
{
    bodydynamics::fti::compose(_delta1, _delta2, _dt2, _delta1_plus_delta2);
}

void ProcessorForceTorqueInertialDynamics::deltaPlusDelta(const Eigen::VectorXd& _delta1,
                                                          const Eigen::VectorXd& _delta2,
                                                          const double           _dt2,
                                                          Eigen::VectorXd&       _delta1_plus_delta2,
                                                          Eigen::MatrixXd&       _jacobian1,
                                                          Eigen::MatrixXd&       _jacobian2) const
{
    bodydynamics::fti::compose(_delta1, _delta2, _dt2, _delta1_plus_delta2, _jacobian1, _jacobian2);
}

void ProcessorForceTorqueInertialDynamics::statePlusDelta(const VectorComposite& _x,
                                                          const Eigen::VectorXd& _delta,
                                                          const double           _dt,
                                                          VectorComposite&       _x_plus_delta) const
{
    auto     x            = _x.vector("PVLO");
    VectorXd x_plus_delta = bodydynamics::fti::composeOverState(x, _delta, _dt);
    _x_plus_delta         = VectorComposite(x_plus_delta, "PVLO", {3, 3, 3, 4});
}

Eigen::VectorXd ProcessorForceTorqueInertialDynamics::correctDelta(const Eigen::VectorXd& _delta_preint,
                                                                   const Eigen::VectorXd& _delta_step) const
{
    return fti::plus(_delta_preint, _delta_step);
}

VectorXd ProcessorForceTorqueInertialDynamics::getCalibration(const CaptureBaseConstPtr _capture) const
{
    if (_capture)
    {
        VectorXd        calibration(13);
        const Vector6d& I_calib = _capture->getStateBlock('I')->getState();
        const Vector3d& C_calib = _capture->getStateBlock('C')->getState();
        const Vector3d& i_calib = _capture->getStateBlock('i')->getState();
        const Vector1d& m_calib = _capture->getStateBlock('m')->getState();
        calibration << I_calib, C_calib, i_calib, m_calib;
        return calibration;
    }
    else
    {
        VectorXd        calibration(13);
        const Vector6d& I_calib = getSensor()->getStateBlockDynamic('I')->getState();
        const Vector3d& C_calib = getSensor()->getStateBlockDynamic('C')->getState();
        const Vector3d& i_calib = getSensor()->getStateBlockDynamic('i')->getState();
        const Vector1d& m_calib = getSensor()->getStateBlockDynamic('m')->getState();
        calibration << I_calib, C_calib, i_calib, m_calib;
        return calibration;
    }
}

bool ProcessorForceTorqueInertialDynamics::voteForKeyFrame() const
{
    // time span
    if (getBuffer().back().ts_ - getBuffer().front().ts_ > params_force_torque_inertial_dynamics_->max_time_span)
    {
        WOLF_DEBUG("PM ", getName(), " vote: time span");
        return true;
    }
    // buffer length
    if (getBuffer().size() > params_force_torque_inertial_dynamics_->max_buff_length)
    {
        WOLF_DEBUG("PM ", getName(), " vote: buffer length");
        return true;
    }
    // dist_traveled
    VectorComposite X0 = getOrigin()->getFrame()->getState();
    VectorComposite X1;
    double          dt = getBuffer().back().ts_ - getOrigin()->getFrame()->getTimeStamp();
    statePlusDelta(X0, getBuffer().back().delta_integr_, dt, X1);
    double dist = (X1.at('P') - X0.at('P')).norm();
    if (dist > params_force_torque_inertial_dynamics_->dist_traveled)
    {
        WOLF_DEBUG("PM ", getName(), " vote: distance traveled");
        return true;
    }
    // angle turned
    double angle = 2.0 * acos(getMotion().delta_integr_(15));
    if (angle > params_force_torque_inertial_dynamics_->angle_turned)
    {
        WOLF_DEBUG("PM ", getName(), " vote: angle turned");
        return true;
    }

    return false;
}

} /* namespace wolf */

#include <core/processor/factory_processor.h>
namespace wolf
{
WOLF_REGISTER_PROCESSOR(ProcessorForceTorqueInertialDynamics);
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorForceTorqueInertialDynamics);
}  // namespace wolf